;
; Copyright 2005, 2006 University of Leiden.
;
; This file is part of MIA+EWS.
;
; MIA+EWS is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; MIA+EWS is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with MIA; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;********************************************************
;#class#
;midiFringe
;#description#
;class of fringe detection and estimation algorithms
;#end_class#
;*******************************************************

PRO midiFringe__Define
COMPILE_OPT idl2,HIDDEN
;********************************************************
;#structure#
; midiFringeStruct
;#inheritance#
; NONE
;#description#
;internal variables for midiFringe object
;#structure_text#
   fringe={midiFringe, $
   fitsFile:'', $;input fits file name
   fringeMaskFile:'', $;input fits mask name
   filterFile:'', $;spectral bandpass
   sensitivityFile:'', $;MIDI spectral sensitivity file
   compressedFile:'', $;compressed version of fits file
   rawData:PTR_NEW(),     $;raw data, if any
   compressedData:PTR_NEW(), $ ;compressed data
   filteredData:PTR_NEW(), $ ;filtered data
   nData:0L, $ ;number of frames
   nScan:0L, $ ;number of scans
   scanLength:0L, $ ;number of frames per scan
   i1:0, $; 1st interferometric channel
   i2:0, $; 2nd interferometric channel
   tZero:1.0, $ time constant for removing low freq trends
   flipOpd:0,  $; VLT opd has wrong sign
   sensitivity:PTR_NEW(), $; instrumental sens as fn of wavenumber
   starSpec:PTR_NEW(), $; spectral power of model star
   scanOpd:PTR_NEW(), $; opd estimated from data
   power:PTR_NEW(), $; spectral power or otherwise
   outputOpd:PTR_NEW(), $; opdRange after filtering
   outType:'', $; output type of filtering operation
   gridPar:fltarr(3), $; gridding data for fft
   dit:0.0, $; exposure time
   grism:'', $; dispersing element
   doPhoto:0, $; normalize plots by photometry
   fftObject:OBJ_NEW() $; fft object
   }
;#end_structure#
;********************************************************
END

PRO midiFringe::save, filename
;************************************************************
   fitsfile=self.fitsFile
   fringeMaskFile=self.fringemaskfile
   filterFile=self.filterfile
   sensitivityFile=self.sensitivityfile
   compressedFile=self.compressedFile
   if (PTR_VALID(self.rawData)) then rawData=*self.rawData else $
      rawData = 0
   if (PTR_VALID(self.CompressedData)) then $
      compressedData=*self.compressedData else compressedData=0
   if (PTR_VALID(self.filteredData)) then filteredData=*self.filteredData $
      else filteredData = 0
   nData=self.ndata
   nScan=self.nscan
   scanLength=self.scanlength
   i1=self.i1
   i2=self.i2
   tZero=self.tzero
   flipOpd=self.flipopd
   if (PTR_VALID(self.sensitivity)) then sensitivity=*self.sensitivity $
      else sensitivity = 0
   if (PTR_VALID(self.starSpec)) then starSpec=*self.starspec else $
      starSpec = 0
   if (PTR_VALID(self.scanOpd)) then scanOpd=*self.scanopd else $
      scanOpd = 0
   if (PTR_VALID(self.power)) then power=*self.power else power=0
   if (PTR_VALID(self.outputOpd)) then outputOpd=*self.outputopd $
      else outputOpd = 0
   outType=self.outtype
   gridPar=self.gridpar
   dit=self.dit
   grism=self.grism
   doPhoto=self.dophoto
   save,file=filename+'_midiFringe', fitsfile, fringeMaskFile, $
      filterFile,sensitivityFile, compressedFile, rawData, $
      compressedData, filteredData, nData, nScan, scanLength, $
      i1, i2, tZero, flipOpd, sensitivity, starspec, scanOpd, $
      power, outputOpd, outType, gridpar, dit, grism, doPhoto
RETURN
END

PRO dispersedFringe__Define
COMPILE_OPT idl2,HIDDEN
;********************************************************
;#structure#
; dispersedFringeStruct
;#inheritance#
; midiFringe object
;#description#
;internal variables for dispersed Fringe object
;#structure_text#
   fringe={dispersedFringe, $
   INHERITS midiFringe, $
   freqFile:'', $; file supplying freq data, otherwise use fitsFile
   nFreq:0L, $
   frameSelect:PTR_NEW(), $
   waveNo:PTR_NEW(), $
   waveLength:PTR_NEW(), $
   medianOpd:0., $
   interfData:PTR_NEW(), $
   rotatedData:PTR_NEW(), $
   autoCorr:PTR_NEW(), $
   delayData:PTR_NEW() $
   }
;#end_structure#
;********************************************************
END

PRO dispersedFringe::save, filename
;************************************************************
   self->midiFringe::save, filename
   freqFile = self.freqFile
   nFreq = self.nFreq
   if (PTR_VALID(self.frameSelect)) then frameSelect = *self.frameSelect $
      else frameSelect = 0
   if (PTR_VALID(self.waveNo)) then waveNo = *self.waveNo $
      else waveNo = 0
   if (PTR_VALID(self.waveLength)) then waveLength = *self.waveLength $
      else waveLength = 0
   medianOpd = self.medianOpd
   if (PTR_VALID(self.interfData)) then interfData = *self.interfData $
      else interfData = 0
   if (PTR_VALID(self.rotatedData)) then rotatedData = *self.rotatedData $
      else rotatedData = 0
   if (PTR_VALID(self.autoCorr)) then autoCorr = *self.autoCorr $
      else autoCorr = 0
   if (PTR_VALID(self.delayData)) then delayData = *self.delayData $
      else delayData = 0
   save,file=filename+'_dispersedFringe', freqFile, nFreq, $
      frameSelect, waveNo, waveLength, interfData, rotatedData, $
      autoCorr, delayData, medianOpd
RETURN
END

FUNCTION dispersedFringe::offsetOpd
RETURN,self.medianOpd
END

PRO scannedFringe__Define
COMPILE_OPT idl2,HIDDEN
;********************************************************
;#structure#
; ScannedFringeStruct
;#inheritance#
; midiFringe object
;#description#
;internal variables for Scanned Fringe object
;#structure_text#
fringe={scannedFringe, $
INHERITS midiFringe, $
fringe:PTR_NEW(), $
filteredFringe:PTR_NEW(), $
kData:PTR_NEW(), $
photo:fltarr(2), $; photometric data, [1ab, 2ab]
maxAmp:PTR_NEW(), $; maximum amplitude per scan
maxOPD:PTR_NEW(), $; opd Positions at max amp
starModelFringe:PTR_NEW(), $; star spectral power as scanned
starMaxAmp:0., $;
scanSelect:PTR_NEW()   $; valid scan flags
}
;#end_structure#
;********************************************************
END

FUNCTION midiFringe::init,file, flipOpd=flipOpd, $
   tZero=tZero, fitsFile = fitsFile, restore=restore, iErr=iErr
;********************************************************
;#function#
;init
;#description#
;create midiFringe object
;store file names and read out some of the header information
;#call#
;fringeObject=OBJ_NEW('midiFringe', fitsFile, maskFile)	
;#inputs#
;\anArg{fitsFile}{string }{filename space-separated set of input files}
;\anArg{maskFile}{string}{filename containing image weighting mask}
;#return#
;\anArg{-}{object}{required object}
;#end_function#
;*******************************************************
;
   if (NOT KEYWORD_SET(restore)) then begin
      self.fringeMaskFile = file
      if (NOT KEYWORD_SET(tZero)) then tZero = 1.
      self.tZero = tZero
      if (KEYWORD_SET(flipOpd)) then self.flipOpd = 1 else self.flipOpd = 0
;   figure out scan size from header
      if (KEYWORD_SET(fitsFile)) then begin
         self.fitsFile = fitsFile
         self->getFileData, fitsFile
      endif
      self.doPhoto = 0
   endif else begin
      RESTORE, file+'_midiFringe'
      self.fitsfile=fitsFile
      self.fringeMaskFile=fringemaskfile
      self.filterFile=filterfile
      self.sensitivityFile=sensitivityfile
      self.compressedFile=compressedFile
      self.rawData=PTR_NEW(rawData)
      self.compressedData=PTR_NEW(compressedData)
      self.filteredData=PTR_NEW(filteredData)
      self.nData=self.ndata
      self.nScan=self.nscan
      self.scanLength=self.scanlength
      self.i1=self.i1
      self.i2=self.i2
      self.tZero=self.tzero
      self.flipOpd=self.flipopd
      self.sensitivity=PTR_NEW(sensitivity)
      self.starSpec=PTR_NEW(starspec )
      self.scanOpd=PTR_NEW(scanopd )
      self.outputOpd=PTR_NEW(outputopd)
      self.outType=self.outtype
      self.gridPar=self.gridpar
      self.dit=self.dit
      self.grism=self.grism
      self.doPhoto=self.dophoto
   endelse
RETURN,1
END

PRO midiFringe::getFileData, fitsFile
;
   filename = oirFileList(fitsFile)
   im = OBJ_NEW('imagedata', filename[0], iErr=iErr)
   if (iErr NE 0) then RETURN
   imh = im->head()
   self.scanLength = imh->getpar('maxstep')
;    figure out channel location from OPT1
   prih = im->prihead()
   beam = prih->getpar('INS OPT1')
   self.grism = prih->getpar('INS GRIS')
   self.dit = prih->getpar('DET DIT')
   print,'Disperser: ',self.grism,' DIT ',self.dit
   OBJ_DESTROY,im
   if (beam EQ 'HIGH_SENS') then begin 
      self.i1 = 0 
      self.i2 = 1 
   endif else if (beam EQ 'SCI_PHOT') then begin
      self.i1=1 
      self.i2 = 2
   endif
RETURN
END

PRO midiFringe::cleanup
COMPILE_OPT idl2,HIDDEN
   PTR_FREE,self.rawData
   PTR_FREE,self.compressedData
   PTR_FREE,self.filteredData
   PTR_FREE,self.scanOpd
   PTR_FREE,self.power
   PTR_FREE,self.starSpec
   PTR_FREE,self.sensitivity
   PTR_FREE,self.outputOpd
   OBJ_DESTROY,self.fftObject
RETURN
END

FUNCTION midiFringe::localOpd ,scan = scan
COMPILE_OPT idl2,HIDDEN
;********************************************************
;#function#
;localOpd
;#description#
;return local opd (due to midi piezos) of this dataset
;in microns.  Optionally break this up into scans (2-d array)
;#call#
;localOpd = midiFringeObject->localOpd(/scan)
;#inputs#
;\anArg{scan}{optional int }{if present, format output data in scans}
;#return#
;\anArg{-}{fltarr}{local opd data}
;#end_function#
;*******************************************************
   opd = (*self.compressedData).localopd[1] - $
      (*self.compressedData).localopd[0]
   opd = 1.e6*opd
   if (KEYWORD_SET(scan)) then RETURN, REFORM(opd, self.scanLength, $
      self.nScan,/overwrite) $
   else RETURN, REFORM(opd, /overwrite)
END

FUNCTION midiFringe::vltiOpd, scan = scan 
COMPILE_OPT idl2,HIDDEN
;********************************************************
;#function#
;vltiOpd
;#description#
;return external opd (due to vlti delay lines) of this dataset
;in microns.  Optionally break this up into scans (2-d array)
;#call#
;delayOpd = midiFringeObject->vltiOpd(/scan)
;#inputs#
;\anArg{scan}{optional int }{if present, format output data in scans}
;#return#
;\anArg{-}{fltarr}{external opd data}
;#end_function#
;*******************************************************
   if (self.flipOpd EQ 0) then $
      opd =  (*self.compressedData).opd[1] - (*self.compressedData).opd[0] $
   else opd = (*self.compressedData).opd[0] - $
            (*self.compressedData).opd[1] 
   opd = 1.e6*opd
   if (KEYWORD_SET(scan)) then RETURN, REFORM(opd, self.scanLength, $
      self.nScan,/overwrite) $
   else RETURN, REFORM(opd,/overwrite)
END

FUNCTION midiFringe::fftObject & RETURN,self.fftObject&END

FUNCTION midiFringe::opd, scan = scan
;
;********************************************************
;#function#
;Opd
;#description#
;return total opd (midi piezos+VLTIDelaylines) of this dataset
;in microns.  Optionally break this up into scans (2-d array)
;#call#
;Opd = midiFringeObject->Opd(/scan)
;#inputs#
;\anArg{scan}{optional int }{if present, format output data in scans}
;#return#
;\anArg{-}{fltarr}{opd data}
;#end_function#
;*******************************************************
   myopd = self->localopd(scan=scan) + self->vltiOpd(scan=scan)
   if (TOTAL(0 EQ FINITE(myopd)) NE 0) then $
      myopd(WHERE(0 EQ FINITE(myopd))) = MEDIAN(myopd)
   if (KEYWORD_SET(scan)) then RETURN, REFORM(myopd, self.scanLength, $
      self.nScan,/overwrite) $
   else RETURN, REFORM(myopd)
END

FUNCTION midiFringe::time, scan=scan, zero=zero
;
;********************************************************
;#function#
;time
;#description#
;return time of each measurement in this data set (sec)
;Optionally break this up into scans (2-d array) a/o remove zero point
;#call#
;time = midiFringeObject->time()
;scantime = midiFringeObject->time(/scan,/zero)
;#inputs#
;\anArg{scan}{optional int }{if present, format output data in scans}
;\anArg{zero}{optional int }{if present, subtract initial time}
;#return#
;\anArg{-}{fltarr}{time data}
;#end_function#
;*******************************************************
   time = 86400.*REFORM((*self.compressedData).time,/overwrite)
   if (KEYWORD_SET(zero)) then time = time - time[0]
   if (KEYWORD_SET(scan)) then RETURN, REFORM(time, self.scanLength, $
      self.nScan,/overwrite)  $
   else RETURN, time
END

PRO dispersedFringe::cleanup
   self->midiFringe::cleanup
   PTR_FREE,self.waveNo
   PTR_FREE,self.waveLength
   PTR_FREE,self.rotatedData
   PTR_FREE,self.interfData
   PTR_FREE,self.delayData
   PTR_FREE,self.frameSelect
   PTR_FREE,self.autoCorr
RETURN
END

PRO scannedFringe::cleanup
   self->midiFringe::cleanup
   PTR_FREE,self.fringe
   PTR_FREE,self.filteredFringe
   PTR_FREE,self.maxAmp
   PTR_FREE,self.maxOPD
   PTR_FREE,self.kData
   PTR_FREE,self.starModelFringe
   PTR_FREE,self.scanSelect
RETURN
END

FUNCTION dispersedFringe::init, file, flipOpd = flipOpd, tZero = tZero, $
   restore=restore, iErr=iErr
   i = self->midiFringe::init(file, flipOpd=flipOpd, tZero=tZero, $
      restore=restore)
   if (KEYWORD_SET(restore)) then begin
      RESTORE,file+'_dispersedFringe'
      self.freqFile = freqFile
      self.nFreq = nFreq
      self.frameSelect = PTR_NEW(frameSelect)
      self.waveNo = PTR_NEW(waveNo)
      self.waveLength = PTR_NEW(waveLength)
      self.medianOpd = medianOpd
      self.interfData = PTR_NEW(interfData)
      self.rotatedData = PTR_NEW(rotatedData)
      self.autoCorr = PTR_NEW(autoCorr)
      self.delayData = PTR_NEW(delayData)
   endif
RETURN,1
END

PRO dispersedFringe::getFitsData, fitsFile, freqFile
   if (N_PARAMS() GT 1) then self.freqFile = freqFile $
   else begin
      z = STRSPLIT(fitsFile,' ',/EXTRACT)
      self.freqFile=z[0]
   endelse
   self.compressedData = PTR_NEW(oirCompressDataC(fitsFile, $
      self.fringeMaskFile, nDim=1, iErr=iErr))
   if (iErr NE 0) then RETURN
   self.nData = N_ELEMENTS(*self.compressedData)
   self.waveNo = PTR_NEW(oirGetWaveNo(self.freqFile))
   self.waveLength = PTR_NEW(2*!pi/(*self.waveNo))
RETURN
END

PRO dispersedFringe::tvRaw, scale=scale, shrink=shrink
      yMag = (!d.Y_size/dSize[1]) > 1
   data = transpose((*self.interfData), [1,0])
   if (KEYWORD_SET(shrink)) then begin
      dSize=SIZE(data,/dim)
      xSize = !d.X_size
      if (xSize LT dSize[0]) then begin
         x = fix(dSize[0]/xSize) 
         xx = x * xSize - 1
         data = REBIN(data(0:(xx-1), *), xSize, dSize[1])
      endif
   endif
   if (NOT KEYWORD_SET(scale)) then tvscl, data $
      else tvm,128+scale*data
RETURN
END

PRO dispersedFringe::formFringes, xsmooth
   if (N_PARAMS() LT 1) then xsmooth = 50
   PTR_FREE,self.interfData
   self.interfData = PTR_NEW(((*self.compressedData).data1 - $
      (*self.compressedData).data2)/self.dit)
   iSize = size(*self.interfData, /dim)
   for i=0, iSize[0]-1 do (*self.interfData)[i,*] = (*self.interfData)[i,*] - $
      fsmooth1(REFORM((*self.interfData)[i,*]), xsmooth)
RETURN
END

PRO dispersedFringe::tvRotated, scale=scale
   IF (NOT PTR_VALID(self.rotatedData)) then print,'Data has not yet been rotated '$
   else if (NOT KEYWORD_SET(scale)) then tvsclm,float(*self.rotatedData) $
   else tvm,128+scale*float(*self.rotatedData)
RETURN
END

PRO dispersedFringe::tvDelay, real=real
   IF (NOT  PTR_VALID(self.delayData)) then begin
      print,'No Delay Data Yet' 
      RETURN
   endif
   if (NOT KEYWORD_SET(scale)) then $
      if (NOT KEYWORD_SET(float)) then tvscl,ABS(*self.DelayData) $
         else tvscl,FLOAT(*self.DelayData)  $
   else if (NOT KEYWORD_SET(float)) then tv,scale*ABS(*self.DelayData) $
      else tv,128+scale*FLOAT(*self.DelayData)  
RETURN
END

PRO dispersedFringe::rotate
   PTR_FREE,self.rotatedData
   opd = self->opd()
;  self.medianOpd = median(opd) ; make opds a reasonably small number
   self.medianOpd = opd[N_ELEMENTS(opd)-50] ; make opds a small number
   opd = opd - self.medianOpd
   self.rotatedData = PTR_NEW(midiShiftSpecFringe((*self.interfData), opd, $
      *self.waveNo))
RETURN
END

PRO dispersedFringe::groupRotate,nMedian
   if (N_PARAMS() LT 1) then nMedian = 10
   opd = median(self->groupOpd(),nMedian) - self.medianOpd
   rotatedData = TRANSPOSE(*self.rotatedData,[1,0])
  rotatedData = midiShiftSpecFringe(rotatedData, -opd, *self.waveNo)
   *self.rotatedData = rotatedData
RETURN
END

PRO dispersedFringe::dePhase, nSmooth=nSmooth, weight=weight
   if (NOT KEYWORD_SET(nSmooth)) then nSmooth = 20
   rotatedData = *self.rotatedData
   s = SIZE(rotatedData, /dim)
   if (NOT KEYWORD_SET(weight)) then weight = REPLICATE(1., s[1])
;smooth rotated data to supress wrong sidelobe and compute average phase
   for i=0, s[1]-1 do rotatedData(*,i)=csmooth1(rotatedData(*,i),nsmooth)
   pave = FLTARR(s[0])
   for i=0, s[0]-1 do pave[i]  = midiPhase(TOTAL(rotatedData(i,*)*weight))
;restore unsmoothed data and remove phase
   rotatedData = *self.rotatedData
    for i=0, s[1]-1 do rotatedData(*,i) = rotatedData(*,i) * exp(-complex(0,1)*pave)
; store
   *self.rotatedData = rotatedData
RETURN
END


FUNCTION dispersedFringe::rawData
   RETURN, *self.interfData
END
FUNCTION dispersedFringe::rotatedData
   RETURN, *self.rotatedData
END

FUNCTION dispersedFringe::coherentFringe
   select = *self.frameSelect
   rd = (*self.rotatedData)[WHERE(select),*]
RETURN, TOTAL(rd,1)/TOTAL(select)
END

FUNCTION dispersedFringe::coherentPhase
   vis = self->coherentFringe()
   RETURN,!radeg*atan(imaginary(vis), float(vis))
END

FUNCTION scannedFringe::init, maskFile, $
   flipOpd = flipOpd, tZero=tZero, fitsFile = fitsFile, iErr=iErr
;
;********************************************************
;#function#
;init
;#description#
;create scannedFringe object; a child of midiFringe object
;after initing midiFringe, compress data and store locally
;compute fringe (I1-I2), high pass filter this, divide by exposure time
;#call#
;fringeObject=OBJ_NEW('midiFringe', fitsFile, maskFile)	
;#inputs#
;\anArg{fitsFile}{string }{filename space-separated set of input files}
;\anArg{maskFile}{string}{filename containing image weighting mask}
;\anArg{tZero}{flt}{optional time window (s) for high-pass filter(def=1.0)}
;#return#
;\anArg{-}{object}{required object}
;#end_function#
;*******************************************************
   iErr = 0
   i = self->midiFringe::init(maskFile, fitsFile=fitsFile, $
      flipOpd=flipOpd, tZero=tZero)
   self.photo= [0.5, 0.5]
   self.starMaxAmp = 1.
   if (KEYWORD_SET(fitsFile)) then self->getFitsData, fitsFile, $
      iErr = iErr
   if (iErr NE 0) then RETURN,0 else RETURN,1
END

PRO scannedFringe::getFitsData, fitsFile, iErr = iErr
;   
   self.fitsFile = fitsFile
   self.compressedData = PTR_NEW(oirCompressDataC(fitsFile, $
      self.fringeMaskFile, iErr=iErr))
   self.nData = N_ELEMENTS(*self.compressedData)
   self.nScan = self.nData/self.scanLength
; force integral number of scans
   nData = self.scanLength * self.nScan
   if (nData LT self.nData) then begin
      tdata = (*self.compressedData)[0:nData-1]
      PTR_FREE, self.compressedData
      self.compressedData = PTR_NEW(tdata)
      self.nData = nData
   endif
   PRINT,' '
   PRINT,self.nData,' Data Points ',self.nScan, ' Scans'
   dTime = (*self.compressedData)[1].time - (*self.compressedData)[0].time
   dTime = 86400.*dTime
   nZero = self.tZero/dTime
   tags = TAG_NAMES(*self.compressedData)
   dCol = WHERE(STRPOS(tags,'DATA') GE 0)
   PTR_FREE,self.fringe
   fringe=REFORM((*self.compressedData).(dCol[self.i1]) - $
      (*self.compressedData).(dCol[self.i2]))
   fringe[0:1] = MEDIAN(fringe[2:200])
   fringe = fringe - SMOOTH(fringe, nZero, /EDGE_TRUNCATE)
   self.fringe = PTR_NEW(REFORM(fringe, self.scanLength, self.nScan)/self.dit)
   self.scanSelect = PTR_NEW(REPLICATE(1B, self.nScan))
RETURN
END

FUNCTION midiFringe::compressedData & RETURN,*self.compressedData&END

FUNCTION midiFringe::filterData 
   if (OBJ_VALID(self.fftObject)) then RETURN,self.fftObject->filterData()
   print,'no fftobject defined'
RETURN,0
END

FUNCTION midiFringe::fft & RETURN, self.fftObject & END

FUNCTION scannedFringe::rawFringe 
;********************************************************
;#function#
;rawFringe
;#description#
;return stored raw fringe array
;#call#
;fringe = fringeObj->rawFringe()
;#inputs#
;#return#
;\anArg{-}{fltarr}{raw fringes organized as scan array}
;#end_function#
;*******************************************************
   if (self.doPhoto EQ 1) then RETURN,*self.fringe*0.5/TOTAL(self.photo) $
   else RETURN,*self.fringe
END

FUNCTION scannedFringe::photo& RETURN,TOTAL(self.photo)&END

FUNCTION scannedFringe::filteredData 
;********************************************************
;#function#
;filteredData
;#description#
;return stored filtered fringes (fft'd then multiplied by filter fn and
;fft'd back) as 2-d array, nGrid*nScan
;#call#
;filteredfringe = fringeObj->filteredData()
;#inputs#
;#return#
;\anArg{-}{fltarr}{filtered fringes organized as scan array}
;#end_function#
;*******************************************************
   if (self.doPhoto EQ 1) then $
      RETURN,*self.filteredData*0.5/TOTAL(self.photo) $
   else RETURN,*self.filteredData
END
FUNCTION scannedFringe::kSpectra 
;********************************************************
;#function#
;kSpectra
;#description#
;return stored fft data (raw fringes fft'd to k-space)
;#call#
;spectra = fringeObj->kSpectra()
;#inputs#
;#return#
;\anArg{-}{complexarr}{complex spectra organized as scan array}
;#end_function#
;*******************************************************
   if (self.doPhoto EQ 1) then $
      RETURN,*self.kData*0.5/TOTAL(self.photo) $
   else RETURN,*self.kData
END

FUNCTION scannedFringe::maxAmp 
;********************************************************
;#function#
;maxAmp
;#description#
;return peak amplitude of filtered fringes as fun of scan number
;#call#
;amplitude = fringeObj->maxAmp()
;#inputs#
;#return#
;\anArg{-}{fltarr}{peak filtered fringe as fn of scan }
;#end_function#
;*******************************************************
   if (self.doPhoto EQ 1) then $
      RETURN,*self.maxAmp/(2*TOTAL(self.photo)*self.starMaxAmp) $
   else RETURN,*self.maxAmp
END

FUNCTION scannedFringe::maxOpd 
;********************************************************
;#function#
;maxOpd
;#description#
;return totalOPD(micron) at position of peak amplitude of filtered fringes 
; as fun of scan number
;#call#
;opd = fringeObj->maxOpD()
;#inputs#
;#return#
;\anArg{-}{fltarr}{OPD of peak filtered fringe as fn of scan }
;#end_function#
;*******************************************************
RETURN,*self.maxOpd
END
FUNCTION scannedFringe::midScanOpd
   opd = self->opd(/scan)
   nOpd = SIZE(opd)
   RETURN, TOTAL(opd,1)/nOpd[1]
END
FUNCTION midiFringe::kGrid 
;********************************************************
;#function#
;kGrid
;#description#
;spatial frequencies (2pi/wavelength) corresponding to abscissa of
;all spectra returned by fft routines
;#call#
;spFreq = fringeObj->kGrid()
;#inputs#
;#return#
;\anArg{-}{fltarr}{spatial frequency array}
;#end_function#
;********************************************************
   if (OBJ_VALID(self.fftObject)) then RETURN, self.fftObject->kGrid()
   RETURN,0
END

FUNCTION midiFringe::outOpd 
;********************************************************
;#function#
;outOpd
;#description#
;OPDs(microns) corresponding to abscissa of all filtered fringes
; returned by FFT routines
;#call#
;spFreq = fringeObj->outOpd()
;#inputs#
;#return#
;\anArg{-}{fltarr}{OPD array}
;#end_function#
;********************************************************
  if (PTR_VALID(self.outputOpd)) then RETURN,*self.outputOpd else RETURN,0 
END

FUNCTION midiFringe::sensitivity 
   if (PTR_VALID(self.sensitivity)) then RETURN,*self.sensitivity
   print,' sensitivity not defined'
   RETURN,0
END

FUNCTION midiFringe::starSpec 
   if (PTR_VALID(self.starSpec)) then RETURN,*self.starSpec
   print,' starSpec not defined'
   RETURN,0
END

FUNCTION midiFringe::starFringe 
   if (PTR_VALID(self.starModelFringe)) then RETURN,*self.starModelFringe
   print,' starModelFringe not defined'
   RETURN,0
END
FUNCTION midiFringe::starAmp 
   RETURN,self.starMaxAmp
END

PRO scannedFringe::set, tZero=tZero, sensFile=sensFile, photo=photo, $
  doPhoto=doPhoto
COMPILE_OPT HIDDEN
   if(KEYWORD_SET(tZero)) then self.tZero = tZero
   if (KEYWORD_SET(photo)) then begin
      self.photo = photo
   endif
   if (N_ELEMENTS(doPhoto) GT 0) then begin
      self.doPhoto = doPhoto
   endif
   if (KEYWORD_SET(sensFile)) then begin
      self.sensitivityFile = sensfile
      if (OBJ_VALID(self.fftObject)) then begin
         sens = OBJ_NEW('filterdata', sensFile)
         kGrid = ABS(self.fftObject->kGrid())
         PTR_FREE,self.sensitivity
         PTR_FREE,self.starSpec
         self.sensitivity=PTR_NEW(sens->interpolate(1.e6*kGrid)) 
         self.starSpec=PTR_NEW((*self.sensitivity)*$
            (abs(kGrid)>.1)^(-0.05))
;           (abs(kGrid)>.1)^(-1.0))
;        self.starSpec=PTR_NEW((*self.sensitivity))
         *self.starSpec = (*self.starSpec)/max(*self.starSpec)
         OBJ_DESTROY,sens
      endif
;    generate trial opd with zero at center
;    then generate trial scan
      opd = self->opd(/scan)
      opd = opd(*,0)
      mOpd = mean(opd)
      opd = opd - mOpd
      kGrid = self->kGrid()
      nOpd = N_ELEMENTS(opd)
      PTR_FREE,self.starModelFringe
      fd = self->filterData()
      fd = .5*(fd + REVERSE(fd))
;     filteredSpec = (*self.starSpec)*fd
      filteredSpec = (*self.starSpec)
      fringe = 0*opd
      for i=0,nOpd-1 do fringe[i] = $
         TOTAL(filteredSpec*COS(opd[i] * kGrid))
      fringe = fringe - mean(fringe)
      fringe = fringe/MAX(fringe)
      self.starModelFringe = PTR_NEW(fringe)
      self.starMaxAmp=MAX(self.fftObject->filter(opd+mOpd, fringe))
   endif
RETURN
END

FUNCTION scannedFringe::scanPower, sq=sq
;********************************************************
;#function#
;scanPower
;#description#
;return array of spectral power per scan integrated over range
; given by internal filter
;#call#
;power =  = scanObject->scanPower() ; returns integrated amplitude
;power =  = scanObject->scanPower(/sq) ; returns integrated amplitude^2
;#inputs#
;\anArg{sq}{flag}{Square output before integrating over band}
;#return#
;\anArg{-}{fltarr}{power as fn of scan number}
;#end_function#
;********************************************************
;
   if (NOT PTR_VALID(self.kData)) then self->kData
   filterData = self.fftObject->filterData()
   power = FLTARR(self.nScan)
   norm = 1.
   if (KEYWORD_SET(phot)) then norm = 0.5/(SQRT(photom[0]*photom[1])+$
      SQRT(photom[2]*photom[3]))
   if (KEYWORD_SET(sq)) then norm = norm^2
   if (KEYWORD_SET(sq)) then $
      for i=0, self.nScan-1 do power(i) = TOTAL(filterdata* $
      (ABS((*self.kData)[*, i]))^2) $
   else  $
      for i=0, self.nScan-1 do power(i) = TOTAL(filterdata* $
      ABS((*self.kData)[*, i])) 
   power = power * norm
RETURN, power
END

PRO scannedFringe::selectPower, lowFraction, highFraction, bAnd=bAnd, bOr=bOr, bNot=bNot
;********************************************************
;#procedure#
;selectPower
;#description#
;set scan select flag according to scan amplitude relative to max.
;#call#
;fringeObj->selectPower,minfraction_of_max, maxFraction of max
;fringeObj->selectPower,fraction_of_max,/bOr...
;#inputs#
;\anArg{minFraction}{flt}{Scans whose average amplitude is >fraction are selected;def=0.}
;\anArg{maxFraction}{flt}{only Scans whose average amplitude is <fraction are selected;def=1.}
;\anArg{bAnd}{flag}{if set, logical AND current selections with power selection}
;\anArg{bOr}{flag}{if set, logical OR current selections with power selection}
;\anArg{bNot}{flag}{if set, logical NAND current selections with power selection}
;\anArg{-}{flag}{if no boolean flags set, power selection replaces current selection}
;#end_procedure#
;********************************************************
   if (N_PARAMS() EQ 0) then lowFraction = 0.
   if (N_PARAMS() LT 2) then highFraction = 1.
   power = self->scanPower()
   mPower = MAX(power)
   sPower = (power GE (lowfraction*mPower)) AND (power LE (highFraction*mPower))
   if (KEYWORD_SET(bAnd)) then *self.scanSelect = $
      (*self.scanSelect) AND sPower $
   else if (KEYWORD_SET(bOr)) then *self.scanSelect =$
      (*self.scanSelect) OR sPower $
   else if (KEYWORD_SET(bNot)) then *self.scanSelect =$
      (*self.scanSelect) AND (NOT sPower) $
   else (*self.scanSelect) = sPower
RETURN
END

PRO scannedFringe::selectTime, bTime, eTime, bAnd=bAnd, bOr=bOr, bNot=bNot
;********************************************************
;#procedure#
;selectTime
;#description#
;Select valid scans by time interval
;#call#
;fringeObj->selectTime,beginTime, endTime
;fringeObj->selectTime,beginTime, endTime, /band...
;#inputs#
;\anArg{bTime}{flt}{Start time of selected data, seconds since beginning of obs.}
;\anArg{eTime}{flt}{End time of selected data, seconds since beginning of obs.}
;\anArg{bAnd}{flag}{if set, logical AND current selections with power selection}
;\anArg{bOr}{flag}{if set, logical OR current selections with power selection}
;\anArg{bNot}{flag}{if set, logical NAND current selections with power selection}
;#end_procedure#
;********************************************************
   time   = self->time(/scan)
   time   = REFORM(TOTAL(time,1)/self.scanLength)
   time   = time - time[0]
   if (N_PARAMS() EQ 0) then bTime = 0.
   if (N_PARAMS() LT 2) then eTime = MAX(time)
   sTime = (time GE bTime) AND (time LE eTime)
   if (KEYWORD_SET(bAnd)) then *self.scanSelect = $
      (*self.scanSelect) AND sTime $
   else if (KEYWORD_SET(bOr)) then *self.scanSelect =$
      (*self.scanSelect) OR sTime $
   else if (KEYWORD_SET(bNot)) then *self.scanSelect =$
      (*self.scanSelect) AND (NOT sTime) $
   else (*self.scanSelect) = sTime
RETURN
END

PRO scannedFringe::autoSelect
;
;    drop max amp below noiseCut * noise(10 scans)
   noiseCut = 3.
;    drop this many points before/after otherwise flagged point
   sideDrop     = 1
;    drop points where peak is near side of window
   sideOpd   = 8.
;    get list of max amplitudes and max opd
   opd = self->opd(/scan)
   fopd = self->maxOpd()
   famp = self->maxAmp()
   namp = N_ELEMENTS(famp)
   select = BYTARR(namp)
;    reject peaks near edge of window
   for i=0,namp-1 do select[i] = (fopd[i] GT (MIN(opd[*,i])+sideOpd)) $
      AND (fopd[i] LT (MAX(opd[*,i])-sideOpd))
    
;    sort to find 95th percentile
   samp = famp(SORT(famp))
   amax = samp[namp-1]
   a95  = samp[.95*namp]
;    mean noise of scans 5-15
   a0   = MEAN(famp[5:15])
;    initial select, above noise and less than 3 * 95th percentil
   select = select AND (famp GT (noiseCut * a0)) AND (famp LT (3*a95))
;    cell size for histogram
   da = .02 * a95
;    no of cells in histogram
   na = 1+a95/da
;    coordinates of histogram
   ax = da * (findgen(na)-.5)
;    histogram itself of unflagged data
   ha = 0*ax
   for i=0,namp-1 do if (select[i]) then begin
      j = famp[i]/da < (na-1)
      ha[j] = ha[j] + 1
   endif
;    natural log of histogram
   la = alog(ha > 1)
;    weight array, intially everything OK
   wa=replicate(1., na)
;    fit gaussian to histogram and plot
   q = polyfitw(ax, la, wa, 2, ya)
   plot,ax, la
   oplot,ax,ya,thick=2
;     drop points more than 2 sigma from peak and refit.
   wa = wa AND (ya GT (MAX(ya) -2))
   q = polyfitw(ax, la, wa, 2, ya)
   oplot,ax,ya,thick=3
;      drop points more than 3 sigma from newly fitted peak
   ea = ax[min(where(ya GT (MAX(ya)-2.5)))]
   select = select AND (famp GT ea)
;      reject points near sides of scan window
;      copy selection list and flag neighbors of bad points
   s = select
   s[0] = 0
   for i=sideDrop,N_ELEMENTS(s)-1-sideDrop do for j=-sideDrop,sideDrop do $
      s[i] = s[i] AND select[i+j] 
   *self.scanSelect = s

RETURN
END

FUNCTION scannedFringe::select
RETURN,(*self.scanSelect)
END

PRO scannedFringe::newMask, newMaskFile, clip
;********************************************************
;#procedure#
;newMask
;#description#
;create a new pixel mask file for the interferometric channels
; by correlating each pixel with fringe signal (from previous mask)
;#call#
;fringeObj->newMask,newMaskFile
;#inputs#
;\anArg{newMaskFile}{str}{output file name to put mask}
;\anArg{clip}{flt}{only correlate with amplitude>clip*max;def=.1}
;#end_procedure#
;********************************************************
; use correlation with fringe to optimize pixel selection
   if (N_PARAMS() LT 2) then clip = 0.1
   signal = REFORM(*self.fringe, N_ELEMENTS(*self.fringe))
   flag = ABS(signal) GT (clip * MAX(ABS(signal)))
   print,'NEWMASK: Creating correlated image with clip =', clip
   corrMask = oirCorrelateImage(self.fitsFile, signal, flag)
   i1 = self.i1
   i2 = self.i2
; take only  values above 20% of max
   m = corrMask.(i1)
   m = m/MAX(m)
   m(WHERE(M LT 0.2)) = 0.
   corrMask.(i1) = m
   m = -corrMask.(i2)
   m = m/MAX(m)
   m(WHERE(M LT 0.2)) = 0.
   corrMask.(i2) = m
; normalize to same sum
   corrMask.(i2) = corrMask.(i2) * $
      TOTAL(corrMask.(i1))/TOTAL(corrMask.(i2))
;  make maximum near 1.
   m = 0.5*(MAX(corrMask.(i1))+MAX(corrMask.(i2)))
   corrMask.(i1) = corrMask.(i1)/m
   corrMask.(i2) = corrMask.(i2)/m
   print,'NEWMASK:Output new mask into ',newMaskFile
   oirNewData, self.fringeMaskFile, newMaskFile, corrMask
   self.fringeMaskfile = newMaskFile
   self->remask
RETURN
END

PRO scannedFringe::reMask
;********************************************************
;#procedure#
;reMask
;#description#
;apply a newly specified mask file to data specified in internal
; stored data file
;#call#
;fringeObj->reMask
;#inputs#
;#end_procedure#
;********************************************************
; someone changed the mask so recompute all internal values
   print,'Remasking input data'
   PTR_FREE,self.compressedData
   self.compressedData = $
      PTR_NEW(oirCompressDataC(self.fitsFile, self.fringeMaskFile))
; force integral number of scans
   nData = self.scanLength * self.nScan
   if (nData LT self.nData) then begin
      tdata = (*self.compressedData)[0:nData-1]
      PTR_FREE, self.compressedData
      self.compressedData = PTR_NEW(tdata)
      self.nData = nData
   endif
   dTime = (*self.compressedData)[1].time - (*self.compressedData)[0].time
   dTime = 86400.*dTime
   nZero = self.tZero/dTime
   tags = TAG_NAMES(*self.compressedData)
   dCol = WHERE(STRPOS(tags,'DATA') GE 0)
   PTR_FREE,self.fringe
   fringe=REFORM((*self.compressedData).(dCol[self.i1]) - (*self.compressedData).(dCol[self.i2]))
   fringe[0:1] = MEDIAN(fringe[2:200])
   fringe = fringe - SMOOTH(fringe, nZero, /EDGE_TRUNCATE)
   PTR_FREE,self.fringe
   self.fringe = PTR_NEW(REFORM(fringe, self.scanLength, self.nScan)/self.dit)
   self->reFilter
RETURN
END

PRO scannedFringe::reFilter
   print,'reFiltering input data'
   self->filter,self.filterfile
   self->kData
RETURN
END

FUNCTION scannedFringe::averagePower, select=select, sq=sq, photo=photo
;********************************************************
;#function#
;averagePower
;#description#
;Average power spectrum (or amplitude) over all (selected) scans
;#call#
;avSpectrum = fringeObj->averagePower()
;avSpectrum = fringeObj->averagePower(/select, /sq, /photo)
;#inputs#
;\anArg{select}{flag}{apply selection flags before averaging}
;\anArg{sq}{flag}{square amplitude->power before averaging}
;\anArg{photo}{flag}{apply photometric normalization if available}
;#end_procedure#
;********************************************************
;
   norm = 1.
   if (KEYWORD_SET(phot)) then norm = 0.5/(SQRT(photom[0]*photom[1])+$
      SQRT(photom[2]*photom[3]))
   if (KEYWORD_SET(sq)) then norm = norm^2

   if (KEYWORD_SET(sq)) then $
   if (NOT PTR_VALID(self.kData)) then self->kData
   if (NOT KEYWORD_SET(select)) then begin
      if (KEYWORD_SET(sq)) then $
        RETURN, norm*TOTAL((ABS(*self.kData))^2,2)/self.nScan $
      else RETURN, norm*TOTAL(ABS(*self.kData),2)/self.nScan 
   endif
   output = 0*(*self.kData)[*,0]
   tScan  = 0
   for i=0,self.nScan-1 do  if ((*self.scanSelect)[i]) $
      then begin 
         if (KEYWORD_SET(sq)) then $
            output=output + (ABS((*self.kData)[*,i]))^2 $
         else $
            output=output + ABS((*self.kData)[*,i]) 
         tScan = tScan +1
      endif
   if (tScan GT 0) then output = output /tScan
RETURN, output
END

PRO scannedFringe::rawMovie,auto=auto, dt=dt
;********************************************************
;#procedure#
;rawMovie
;#description#
;put a movie of raw scans on tv display
;#call#
;fringeObj->rawMovie
;fringeObj->rawMovie,dt=.05,/auto
;#inputs#
;\anArg{dt}{flt}{pause between scans (s)}
;\anArg{auto}{flag}{if set fix x- and y- scales by min/max over all scans}
;#end_procedure#
;********************************************************
COMPILE_OPT HIDDEN
;
  if (self.doPhoto EQ 1) then norm = 0.5/TOTAL(self.photo) $
     else norm = 1.
  if (NOT PTR_VALID(self.fringe))$
     then begin
        print,'There is no data '
        RETURN
     endif
  if (NOT KEYWORD_SET(dt)) then dt = 0.1
  opd = self->opd(/scan)
  if (KEYWORD_SET(auto)) then begin
     oldxr = !x.range
     oldyr = !y.range
     !x.range = [min(opd), max(opd)]
     !y.range = norm*[min(*self.fringe),max(*self.fringe)]
  endif
  for iScan = 0, self.nScan -1 do begin
     plot, opd(*,iScan), norm*(*self.fringe)[*,iScan]
     wait,dt
  endfor
  if (KEYWORD_SET(auto)) then begin
     !x.range = oldxr 
     !y.range = oldyr
  endif
RETURN
END

PRO scannedFringe::filterMovie,auto=auto, dt=dt
;********************************************************
;#procedure#
;filterMovie
;#description#
;put a movie of filtered scans on tv display; requires filter to be specified
;#call#
;fringeObj->rawMovie
;fringeObj->rawMovie,dt=.05,/auto
;#inputs#
;\anArg{dt}{flt}{pause between scans (s)}
;\anArg{auto}{flag}{if set fix x- and y- scales by min/max over all scans}
;#end_procedure#
;********************************************************
COMPILE_OPT HIDDEN
  if (self.doPhoto EQ 1) then norm = 0.5/(TOTAL(self.photo)*self.starMaxAmp) $
     else norm = 1.
  if ((NOT PTR_VALID(self.outputOpd)) OR (NOT PTR_VALID(self.filteredData))) $
     then begin
        print,'There is no filtered data '
        RETURN
     endif
  if (NOT KEYWORD_SET(dt)) then dt = 0.1
  if (KEYWORD_SET(auto)) then begin
     oldxr = !x.range
     oldyr = !y.range
     !x.range = [min(*self.outputOpd), max(*self.outputOpd)]
     !y.range = norm*[0,max(*self.filteredData)]
  endif
  for iScan = 0, self.nScan -1 do begin
     plot, *self.outputOpd, norm*(*self.filteredData)[*,iScan]
     wait,dt
  endfor
  if (KEYWORD_SET(auto)) then begin
     !x.range = oldxr 
     !y.range = oldyr
  endif
RETURN
END

PRO scannedFringe::kData, nGrid=nGrid,ampfix=ampfix
;********************************************************
;#procedure#
;kData
;#description#
;calculates internal spectral data (after fft from delay->spatial Frequency) on
; per scan array
;#call#
;fringeObj->kData
;fringeObj->kData, nGrid=512
;#inputs#
;\anArg{nGrid}{int}{# points in fft gridding array; default is 1024}
;#end_procedure#
;********************************************************
COMPILE_OPT HIDDEN
;
   if (NOT KEYWORD_SET(nGrid)) then nGrid = self.gridPar[2] 
   if (nGrid LT 32) then nGrid = 1024
   self.gridPar[2] = nGrid
   opd = self->opd(/scan)
   minOpd = min(opd)
   maxOpd = max(opd)
;  gridPar = [minOpd-100, (200+maxOpd-minOpd)/nGrid, nGrid]
   gridPar = [minOpd-100, 1., nGrid]
      OBJ_DESTROY, self.fftObject
      self.fftObject = OBJ_NEW('fft1d', gridPar)
      self.gridPar = gridPar
      PTR_FREE,self.outputOpd
      self.outputOpd = PTR_NEW(self.fftObject->coord())
      if (self.filterFile NE '') then begin
         filter = OBJ_NEW('filterData', self.filterFile, iErr=iErr)
         self.fftObject->set, filter=$
            filter->interpolate(1.e6*self.fftObject->kGrid())
         OBJ_DESTROY, filter
      endif
   PTR_FREE,self.kData
   self.kData = PTR_NEW(complexarr(nGrid,self.nScan))
   if (N_ELEMENTS(ampfix) GT 0) then fixamp=REFORM(ampfix, self.scanLength,self.nScan)
   for iScan = 0, self.nScan-1 do begin
      fringe = (*self.fringe)[*,iScan]
      if (N_ELEMENTS(ampfix) GT 0) then fringe = fringe /fixamp(*,iScan)
      fringe = fringe - mean(fringe)
      (*self.kData)[*,iScan] = self.fftObject->cfft(opd(*,iScan), fringe)
   endfor
RETURN
END

PRO scannedFringe::filter, filterFile, outType=outType, nGrid=nGrid
;********************************************************
;#procedure#
;filter
;#description#
;apply a frequency filter to internal raw fringes
; get filter data from disk, store, apply, calculate max amp/opd
; in filtered data and store
;#call#
;fringeObj->filter, filterFile
;fringeObj->filter, filterFile, outType='COMPLEX' 
;#inputs#
;\anArg{filterFile}{str}{ascii filter file describing band-pass}
;\anArg{outType}{str}{'AMP' (default)/'COMPLEX','REAL';}
;\anArg{-}{-}{Convert output complex fringe to this before storing}
;#end_procedure#
;********************************************************
COMPILE_OPT HIDDEN
; check defaults first filter
   if (N_PARAMS() LT 1) then filterFile = ''
   sameFilter = (self.filterFile EQ filterFile)
   self.filterFile = filterFile
; check outType
   if (NOT KEYWORD_SET(outType)) then outType = self.outType
   sameOutType = (outType EQ self.outType)
   if (self.outType EQ '') then self.outType = 'AMP'
; check nGrid
   if (NOT KEYWORD_SET(nGrid)) then nGrid = self.gridPar[2] 
   if (nGrid LT 32) then nGrid = 1024
   self.gridPar[2] = nGrid
; compute instrumental opd
   opd = self->opd(/scan)
   if (TOTAL(FINITE(opd) NE 1) NE 0) then $
      opd(WHERE(FINITE(opd) NE 1)) = MEDIAN(opd)
; compute gridding parameters for FFTs
;   gridPar = [min(opd), (max(opd)-min(opd))/nGrid, nGrid]
   minOpd = min(opd)
   maxOpd = max(opd)
;  gridPar = [minOpd-100, (200+maxOpd-minOpd)/nGrid, nGrid]
   gridPar = [minOpd-100, 1., nGrid]
   OBJ_DESTROY, self.fftObject
   self.fftObject = OBJ_NEW('fft1d', gridPar)
   self.gridPar = gridPar
   PTR_FREE,self.outputOpd
   self.outputOpd = PTR_NEW(self.fftObject->coord())
   if (self.filterFile NE '') then begin
      filter = OBJ_NEW('filterData', self.filterFile )
      if (NOT OBJ_VALID(filter)) then begin
            ewsPrintErrMsg
            RETURN
      endif
      self.fftObject->set, filter=$
         filter->interpolate(1.e6*self.fftObject->kGrid())
      OBJ_DESTROY, filter
   endif else self.fftObject->set, filter=1.0
   if (outType EQ 'REAL') then begin
      self.filteredData = PTR_NEW(fltarr(nGrid, self.nScan))
      for iScan = 0, self.nScan-1 do (*self.filteredData)[*,iScan] = $
         FLOAT(self.fftObject->cFilter(opd(*,iScan), (*self.fringe)[*,iScan]))
   endif else if (self.outType EQ 'COMPLEX') then begin
      self.filteredData = PTR_NEW(complexarr(nGrid, self.nScan))
      for iScan = 0, self.nScan-1 do (*self.filteredData)[*,iScan] = $
         self.fftObject->cFilter(opd(*,iScan), (*self.fringe)[*,iScan])
   endif else begin
      self.outType = 'ABS'
      self.filteredData = PTR_NEW(fltarr(nGrid, self.nScan))
      for iScan = 0, self.nScan-1 do (*self.filteredData)[*,iScan] = $
         ABS(self.fftObject->cFilter(opd(*,iScan), (*self.fringe)[*,iScan]))
   endelse
; compute max signal per scan and its position
   PTR_FREE,self.maxAmp
   PTR_FREE,self.maxOpd
   self.maxAmp = PTR_NEW(fltarr(self.nScan))
   self.maxOpd = PTR_NEW(fltarr(self.nScan))
   for iScan = 0, self.nScan -1 do begin
      (*self.maxOpd)[iScan] = pk(*self.outputOpd, ABS((*self.filteredData)[*,iScan]),z) 
      (*self.maxAmp)[iScan] = z
   endfor
; if sensitivity file known, create model spectra
   if (self.sensitivityFile NE '') then self->set,sensFile=self.sensitivityFile
RETURN
END

FUNCTION scannedFringe::peakFilteredOpd
COMPILE_OPT HIDDEN
   nGrid = self.gridPar[2]
   output = fltarr(self.nScan)
   opd = self->opd()
   oldFilter = self->filterData()
   self.fftObject->set,filter=oldFilter+REVERSE(oldFilter)
   for iScan = 0, self.nScan-1 do begin
      fringe = self.fftObject->filter(opd(*,iScan), (*self.fringe)[*,iScan])
      output[iScan] = pk(*self.outputOpd, ABS(fringe))
   endfor
   self.fftObject->set,filter=oldFilter
RETURN,output
END

FUNCTION scannedFringe::phaseFilterData
COMPILE_OPT HIDDEN
   nGrid = self.gridPar[2]
   output = fltarr(nGrid, self.nScan)
   opd = self->opd(/scan)
   oldFilter = self->filterData()
   self.fftObject->set,filter=oldFilter+REVERSE(oldFilter)
   for iScan = 0, self.nScan-1 do $
      output(*,iScan) = self.fftObject->filter(opd(*,iScan), $
      (*self.fringe)[*,iScan])
   self.fftObject->set,filter=oldFilter
RETURN,output
END

PRO scannedFringe::tvRaw, mag, maxmag=maxmag, scale=scale, title=title
;********************************************************
;#procedure#
;tvRaw
;#description#
;put grey-scale "waterfall" plot of raw fringes on TV
;#call#
;fringeObj->tvRaw, 2
;fringeObj->tvRaw, 2, scale=1.e-3
;#inputs#
;\anArg{mag}{int}{blow up image by this factor before displaying}
;\anArg{scale}{flt}{Multiplication factor for grey scale; if ommited autoscale}
;#end_procedure#
;********************************************************
   data = transpose(*self.fringe, [1,0])
   dSize = SIZE(data,/dim)
   if (N_PARAMS() LT 1) then mag = 1
   xMag = mag
   yMag = mag
   if (KEYWORD_SET(maxmag)) then begin
      xMag = (!d.X_size/dSize[0]) > 1
      yMag = (!d.Y_size/dSize[1]) > 1
   endif
   if (NOT KEYWORD_SET(scale)) then $
      tvscl, REBIN(data, xMag*dSize[0], yMag*dSize[1]) $
      else tv,128+scale*REBIN(data, xMag*dSize[0], xMag*dSize[1])
   for i=0,N_ELEMENTS(*self.scanSelect)-1 do $
      if (NOT (*self.scanSelect)[i]) then plots,xMag*i,0,/dev,psym=3
RETURN
END

PRO scannedFringe::tvRawFilter, mag, scale=scale, group=group, $
   noFilter=noFilter, noclip=noclip, maxplot=maxplot, title=title, clip=clip
;********************************************************
;#procedure#
;tvRawFilter
;#description#
;tv Greyscale (waterfall) of fringes after filtering onto uniform OPD grid
;so that motion of piezos/VLTdelays is also represented
;OPD range is normally clipped to include only region with signal > 10% of pk.
;#call#
;fringeObj->tvRawFilter, mag
;fringeObj->tvRawFilter, 2,scale=.1, /noclip, /noFilter, /group, /maxplot, title='a plot'
;#inputs#
;\anArg{mag}{int}{blow up factor before plotting}
;\anArg{scale}{float}{blow up factor before plotting}
;\anArg{group}{flag}{plot group envelope instead of fringe itself}
;\anArg{noFilter}{flag}{plot apply nop-filter (1.0) instead of stored filter}
;\anArg{noclip}{flag}{show all OPDs instead of only where decent signal}
;\anArg{clip}{flt}{Decent signal is this fraction of peak;default=.1}
;\anArg{maxplot}{flag}{overplot position of max fringe amplitude}
;\anArg{title}{string}{title on top of plot}
;#end_procedure#
;********************************************************
COMPILE_OPT HIDDEN
   if (N_PARAMS() LT 1) then mag = 1
   if (NOT KEYWORD_SET(clip)) then clip = .1
   oldFilter = self->filterData()
   if (KEYWORD_SET(noFilter)) then self.fftObject->set,filter=1.0
   nGrid  = self.gridPar[2]
   output = fltarr(nGrid, self.nScan)
   opd    = self->opd(/scan)
   time   = self->time(/scan)
   time   = REFORM(TOTAL(time,1)/self.scanLength)
   time   = time - time[0]
   fringe = (*self.fringe)
   if (KEYWORD_SET(group)) then offset = 0 else offset = 64.
   for i= 0, self.nScan-1 do fringe(*,i)=fringe(*,i)-mean(fringe(*,i))
   if (KEYWORD_SET(group)) then  for iScan = 0, self.nScan-1 do $
      output(*,iScan) = ABS(self.fftObject->cfilter(opd(*,iScan), $
      fringe[*,iScan])) $
   else  for iScan = 0, self.nScan-1 do $
      output(*,iScan) = self.fftObject->filter(opd(*,iScan), $
      fringe[*,iScan])
   iOpd = 0
   jOpd = N_ELEMENTS(*self.outputOpd)-1
   if (NOT KEYWORD_SET(noclip)) then begin
      maxFringe = MAX(output)
      xFringe = clip * maxFringe
      iOpd = N_ELEMENTS(*self.outputOpd)-1
      jOpd = 0
      for i=0, self.nScan-1 do begin
         oo = output(*,i)
         if (TOTAL(oo GT xFringe) GT 0) then begin
            iOpd = iOpd < MIN(WHERE(oo GT xFringe))
            jOpd = jOpd > MAX(WHERE(oo GT xFringe))
         endif
      endfor
      iOpd = (iOpd - 10)>0
      jOpd = (jOpd + 10)<N_ELEMENTS(*self.outputOpd)-1
      output = output(iOpd:jOpd,*)
   endif
   erase
   if (NOT KEYWORD_SET(scale)) then $
      tvsclm,transpose(output,[1,0]), mag $
      else tvm,offset+scale*transpose(output,[1,0]), mag
   for i=0,N_ELEMENTS(*self.scanSelect)-1 do $
      if (NOT (*self.scanSelect)[i]) then plots,mag*i,0,/dev,psym=1
   xSize = mag*self.nScan
   ySize = mag*(jOpd-iOpd+1)
   if (NOT KEYWORD_SET(title)) then title = ''
   xr = !x.range
   yr = !y.range
   !x.range = 0
   !y.range = 0
   plot,[min(time),max(time)], $
      [(*self.outputopd)[iOpd], (*self.outputOpd)[jOpd]],$
      charsize=.01,/dev, pos=[0,0,xSize,ySize],/nodata,/noerase,$
      xstyle=5,ystyle=5,title='waterfall '+ title
   axis,0, ySize, /dev, xaxis=1,xstyle=1,ystyle=1,xtitle='TIME(S)'
   axis,xSize, 0, /dev, yaxis=1,xstyle=1,ystyle=1,ytitle='OPD(micron)'
   if (KEYWORD_SET(maxplot)) then oplot,time,self->maxopd()
   !x.range = xr
   !y.range = yr
   self.fftObject->set,filter=oldFilter
RETURN
END

PRO scannedFringe::tvPower, mag, scale=scale, sq=sq, title=title
;********************************************************
;#procedure#
;tvPower
;#description#
;greyscale "waterpower" plot of spectral power/amplitude as fn of time
;overplot average power for all scans
;#call#
;fringeObj->tvPower, mag
;fringeObj->tvPower, mag, scale=1.e-6, /sq, title='a plot'
;#inputs#
;\anArg{mag}{int}{blow up factor before plot}
;\anArg{scale}{flt}{override default grey scaling}
;\anArg{sq}{flag}{square amplitudes->power before plot}
;\anArg{title}{string}{title on top of plot}
;#end_procedure#
;********************************************************
;
COMPILE_OPT HIDDEN,idl2
   if (N_PARAMS() LT 1) then mag = 1
   if (NOT KEYWORD_SET(title)) then title = ''
   waveNo = self->kGrid()
   waveZero = MAX(WHERE(waveNo LE 0))
   opd = self->opd()
   dOpd = opd[1] - opd[0]
   dWave = 6.28/dOpd
   deltaWave = waveNo[waveZero] - waveNo[waveZero-1]
   nWave = 0.5*dWAve/deltaWave
   time   = self->time(/scan)
   time   = REFORM(TOTAL(time,1)/self.scanLength)
   time   = time - time[0]
   kData = ABS((*self.kData)[waveZero:waveZero+nWave,*])
   if (KEYWORD_SET(sq)) then kData=kData^2
   kAverage = total(kData,2)/self.nScan
;   kRMS = kAverage
;   for i=0, nWave-1 do kRMS[i]=rms(REFORM(kData[i,*]))
if (NOT KEYWORD_SET(scale)) then $
   tvsclm,ABS(transpose(kData, [1,0])),mag $
   else tvm,scale*transpose(ABS(kData),[1,0]),mag
   for i=0,N_ELEMENTS(*self.scanSelect)-1 do $
      if (NOT (*self.scanSelect)[i]) then plots,mag*i,0,/dev,psym=1
   xSize = mag*self.nScan
   ySize = mag*(nWave+1)
   xr = !x.range
   yr = !y.range
   !x.range = 0
   !y.range = 0
   plot,[min(time),max(time)], $
      [waveNo[waveZero], waveNo[waveZero+nWave]]/(2*!pi),$
      charsize=.01,/dev, pos=[0,0,xSize,ySize],/nodata,/noerase,$
      xstyle=5,ystyle=5, title='POWER '+title
   axis,0, ySize, /dev, xaxis=1,xstyle=1,ystyle=1,xtitle='TIME(s)'
   axis,xSize, 0, /dev, yaxis=1,xstyle=1,ystyle=1,ytitle='K(micron^-1)'
   plot,kAverage,waveNo[waveZero:waveZero+nWave]/(2*!pi), $
   pos=[0,0,xSize,ySize], /noerase,xstyle=5,ystyle=5,/dev
;   oplot,kAverage+kRms,waveNo[waveZero:waveZero+nWave]/(2*!pi), line=2
;   oplot,kAverage-kRms,waveNo[waveZero:waveZero+nWave]/(2*!pi), line=2
   !x.range = xr
   !y.range = yr
RETURN
END

FUNCTION midiFringe::VisStats
   vis = self->maxAmp()
   vis = vis(WHERE(*self.scanSelect GT 0))
RETURN,[MEAN(vis), RMS(vis), RMS(vis)/SQRT(N_ELEMENTS(vis)-1)]
END

FUNCTION dispersedFringe::waveNo
   RETURN,*self.waveNo
END

FUNCTION dispersedFringe::lambda
   RETURN,2*!pi/(*self.waveNo)
END

PRO dispersedFringe::delay, xsmooth = xsmooth
   if (NOT KEYWORD_SET(xsmooth)) then xsmooth=6
   spFreq = self->waveNo()
   rotSize = SIZE(*self.rotatedData,/dim)
   gridData = [min(spFreq)-.1, (.2+max(spFreq)-min(spFreq))/512.,512]
;  set up fourier transform 
   fourier = OBJ_NEW('fft1d', gridData)
   PTR_FREE, self.delayData
   self.delayData = PTR_NEW(COMPLEXARR(rotSize[0], gridData[2]))
;    FFT to delay domain
   for i=0, rotSize[0]-1 do begin
      deZero = REFORM((*self.rotatedData)[i,*])
;     deZero = deZero - mean(deZero)
      (*self.delayData)[i,*] = fourier->cfft(spFreq, deZero )
   endfor
   for i=0,gridData[2]-1 do (*self.delayData)[*,i] = $
      csmooth1((*self.delayData)[*,i],xsmooth)
   PTR_FREE,self.outputOpd
   self.outputOpd = PTR_NEW(fourier->kGrid() + self.medianOpd)
   OBJ_DESTROY, fourier
END

FUNCTION dispersedFringe::delayData
   RETURN,*self.delayData
END

FUNCTION dispersedFringe::groupOpd
   if (NOT PTR_VALID(self.delayData)) then begin
      print,'Delay data not yet calculated'
      RETURN,0
   endif
   gs = SIZE(*self.delayData,/dim)
   gOpd = fltarr(gs[0])
   opd = self->outopd()
   for i=0,gs[0]-1 do gOpd[i] = pk(opd, abs((*self.delayData)[i,*]))
RETURN, gOpd
END


FUNCTION dispersedFringe::groupAmp
   if (NOT PTR_VALID(self.delayData)) then begin
      print,'Delay data not yet calculated'
      RETURN,0
   endif
   gs = SIZE(*self.delayData,/dim)
   gAmp = fltarr(gs[0])
   opd = self->outopd()
   for i=0,gs[0]-1 do begin 
      x = pk(opd, abs((*self.delayData)[i,*]),y) 
      gAmp[i] = y
   endfor
RETURN, gAmp
END


PRO dispersedFringe::autoSelect, deltaOpd = deltaOpd, jumpOpd=jumpOpd, $
   sideDrop = sideDrop, ampCut = ampCut
;
;    drop this many points before/after otherwise flagged point
   if (NOT KEYWORD_SET(sideDrop)) then sideDrop     = 1
;    drop points if there is opd jump
   if (NOT KEYWORD_SET(jumpOpd)) then jumpOpd   = 10.
;    drop points if scan opd differs from true opd
   if (NOT KEYWORD_SET(deltaOpd)) then deltaOpd = 150.
;    drop points in median smoothed amplitude
   if (NOT KEYWORD_SET(ampCut)) then ampCut = .5
;    get list of max amplitudes and max opd
   opd = self->opd()
   fOpd = self->groupOpd()
   fa   = self->groupAmp()
   nOpd = N_ELEMENTS(fopd)
   dOpd = abs(opd - fOpd)
   s = dOpd LT deltaOpd
   jOpd = fopd - .5*(shift(fopd,1)+shift(fopd,-1))
   s = s AND (ABS(jOpd) LT jumpOpd)
   fm = median(fa,200)
   fn = n_elements(fm)
   ffm = MAX(fm(100:fn-100))
   s = s AND (fm GT (ampCut * ffm))
   select = s
   for i=sideDrop,N_ELEMENTS(s)-1-sideDrop do for j=-sideDrop,sideDrop do $
      s[i] = s[i] AND select[i+j] 
   PTR_FREE,self.frameSelect
   self.frameSelect = PTR_NEW(s)
RETURN
END

FUNCTION dispersedFringe::select
   if (NOT PTR_VALID(self.frameSelect)) then begin
      print,'No selection yet made'
      RETURN,0
   endif
RETURN,*self.frameSelect
END

PRO dispersedFringe::autoSpec, noSelect=noSelect, xsmooth=xsmooth
   if (NOT KEYWORD_SET(xsmooth)) then xsmooth = 1.5
   if (NOT PTR_VALID(self.rotatedData)) then begin
      print,'no rotated data present'
      RETURN
   endif
   if (KEYWORD_SET(noSelect) OR (NOT PTR_VALID(self.frameSelect)) ) then $
      data = *self.rotatedData $
     else data = (*self.rotatedData)[WHERE(*self.frameSelect), *]
   dataSize = SIZE(data,/dim)
   PTR_FREE, self.autoCorr
   self.autoCorr = PTR_NEW(FLTARR(dataSize[0],dataSize[1]))
   for i=0, dataSize[1]-1 do begin
      sc = csmooth1(reform(data[*,i]), xsmooth)
      (*self.autoCorr)[*,i]=FLOAT(FFT(ABS(FFT(sc,1))^2,-1))/dataSize[0]
   endfor
RETURN
END

FUNCTION dispersedFringe::autoCorr
   RETURN,*self.autoCorr
END

PRO dispersedFringe::autoPower, cTime
   if (N_PARAMS() LT 1) then cTime = 1.
   time = self->time()
   dTime = time[2] - time[1]
   nFit = FIX(cTime/dTime)
   if (N_PARAMS() LT 1) then nFit = 2
   if (NOT PTR_VALID(self.autoCorr)) then self->autoSpec
   autoSize = SIZE(*self.autoCorr, /dim)
   PTR_FREE, self.power
   self.power = PTR_NEW(FLTARR(autoSize[1]))
   xx =FINDGEN(2*nFit+1)-nFit
   ww = ABS(xx) GE 1
   for i=0, autoSize[1]-1 do begin
      a = SHIFT((*self.autoCorr)(*,i), nFit)
      b=a(0:2*nFit)
      c = POLYFITW(xx,b,ww,2,y)
      (*self.power)[i] =y[nFit]
   endfor
RETURN
END

FUNCTION dispersedFringe::specPower
   if (NOT PTR_VALID(self.power)) then begin
      print,'Spectral power not yet computed'
      RETURN,0
   endif
RETURN,*self.power
END

FUNCTION dispersedFringe::photoPower
   if ((NOT PTR_VALID(self.maskFlux[2,0])) OR $
      NOT PTR_VALID(self.maskFlux[2,1])) then begin
      print,'photo power data not yet available'
      RETURN,0
   endif
RETURN,*self.maskFlux[2,0] + *self.maskFLux[2,1]
END

FUNCTION dispersedFringe::specAmp
   if (NOT PTR_VALID(self.power)) then begin
      print,'Spectral power not yet computed'
      RETURN,0
   endif
RETURN,SQRT(*self.power>0)
END

PRO dispersedFringe::selectTime, bTime, eTime, bAnd=bAnd, bOr=bOr, bNot=bNot
;********************************************************
;********************************************************
   time   = self->time()
   time = time - time[0]
   if (N_PARAMS() EQ 0) then bTime = 0.
   if (N_PARAMS() LT 2) then eTime = MAX(time)
   sTime = (time GE bTime) AND (time LE eTime)
   if (KEYWORD_SET(bAnd)) then *self.frameSelect = $
      (*self.frameSelect) AND sTime $
   else if (KEYWORD_SET(bOr)) then *self.frameSelect =$
      (*self.frameSelect) OR sTime $
   else if (KEYWORD_SET(bNot)) then *self.frameSelect =$
      (*self.frameSelect) AND (NOT sTime) $
   else (*self.frameSelect) = sTime
RETURN
END
