/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#include "minrtsFrameData.h"
#define  DEBUG
#define   deltaY       10
#define  SMOOTHX        0 

/**********************************************************
 * Usage:
 *    oirCrossCoeff AFile BFile coordFile outFile [-sky]
 *
 *    AFiles the output of oirChopPhotoImages for AOPEN photometry on a bright star
 *    BFiles the output of oirChopPhotoImages for BOPEN photometry on a bright star
 *
 *    Output is a IMAGING_DATA table in a fits file with 4 rows
 *    each with only one DATA array 
 *
 *    1:  I1/A
 *    2:  I1/B
 *    3:  I2/A
 *    4:  I2/B
 ************************************************************* */
void Usage() {
   printf("Usage:\n");
   printf("   oirCrossCoeff Afile  BFile coordFile outFile [-sky] \n\n");
   printf("   Afile = output of oirChopPhotoImages for AOPEN photometry files \n");
   printf("   Bfile = output of oirChopPhotoImages for BOPEN photometry files \n");
   printf("   coordFile = MIDI data/detector file containing spectral curvature information\n");
   printf("   outfile = output file name\n");
   printf("   -sky (optional): use sky emission instead of star to estimate coefficients\n");
   printf("  \n");
}

/*
   float pixel copying utility
*/
void pixCpy (float *dest, float* source, long nPix) {
   memcpy (dest, source, nPix*sizeof(float));
   return;
}  /* end pixCpy */

int main (int argc, char** argv) {

   char opt1[12];
   char gris[12];
   char shutter[12];

   int iShutter;
   int i,ierr, iregion;
   int nx, ny;
   long  nPix;
   int ix, iy;
   int iKey;
   float y;

   int chan1, chan2;

   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *outFile = NULL;


   imaging_fdata_table_def *odata;
   imaging_detector_table_def *DetectorDef;
   imaging_detector_table_def *CoordDef;
   imaging_fdata_row *outData;
   int nRegion;


   char *Ashutter = "AOPEN";
   char *Bshutter = "BOPEN";
   int  nShutter;
   char *sciPhot  = "SCI_PHOT";

   char history[80];

   minrtsFrameData *ATarget;
   minrtsFrameData *scratch;
   minrtsFrameData *mask;
   minrtsFrameData *BTarget;
   minrtsFrameData *ACompressed;
   minrtsFrameData *BCompressed;

   int useSky;
   int inRow;
   float *smWork;
   ierr   = 0;
   useSky = 0;
/*
      see if nDrop is specified externally
*/
   if (argc < 5) {
      Usage();
      return 1;
   }

/*
       check for sky option
*/
   if (argc == 6 && (strcmp(argv[5], "-sky") == 0)) useSky = 1;

   if (useSky) inRow = 7; else inRow = 1;
/*
       get detector table with coordinate distortion info
*/
      printf("      ****** oirCrossCoeff *****\n"); 
      printf("   Extracting coordinate data from %s \n", argv[3]);
      i = fits_open_file(&inFile, argv[3], READONLY, &ierr) ;
      if (ierr != 0) {
         printf("cfitsio error %i opening coord file %s\n", ierr, argv[3]);
         printf("%s \n",strerror(errno));
         return 1;
      }
      CoordDef = read_imaging_detector_table(inFile, &ierr);
      if (ierr !=0) {
         printf("error %i opening Coord detector table in %s\n", 
            ierr, argv[3]);
         return 1;
      }
      i = fits_close_file(inFile, &ierr);
/*        
       Loop over input files, print some info
*/
   for (iShutter = 1;iShutter<=2; iShutter++) {

/*
      open first file and get a lot of parameters for iShutter=1
*/
      i = fits_open_file(&inFile, argv[iShutter], READONLY, &ierr) ;
      if (ierr != 0) {
         printf("cfitsio error %i opening file %s\n", ierr, argv[iShutter]);
         printf("%s \n",strerror(errno));
         return 1;
      }
      i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS SHUT NAME", 
         shutter, 0, &ierr);
      printf("shutter = %s\n", shutter);
      nShutter = 0;
      if (strncmp(shutter, Ashutter, 5)==0) nShutter=1;
      if (strncmp(shutter, Bshutter, 5)==0) nShutter=2; 
      if (iShutter == 1) {
         ierr = 0;
         i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", 
            opt1,0, &ierr);
         if (strncmp(opt1, sciPhot,8) ) {
            printf("I dont handle OPT1 = %s\n", opt1);
            return 1;
         }
         i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS GRIS NAME", 
            gris,0,&ierr);
         printf("disperser is = %s\n", gris);
/*
      prepare output file
*/
         strcpy(outFileName, bang);
         strcat(outFileName, argv[4]);

         fits_create_file (&outFile, outFileName, &ierr);
         if (ierr !=0) {
            fprintf (stderr, "error opening fits output %s %i\n",argv[4], ierr);
            printf("%s \n",strerror(errno));
            return 1;
         }
         i = fits_copy_header(inFile, outFile, &ierr);
         sprintf(history,"oirCrossCoeff AFILE %s", argv[1]);
         i = fits_write_history(outFile, history, &ierr);
         sprintf(history,"oirCrossCoeff BFILE %s", argv[2]);
         i = fits_write_history(outFile, history, &ierr);
/*         sprintf(history,"oirCrossCoeff cFILE %s", argv[3]);
         i = fits_write_history(outFile, history, &ierr);
*/

/*
      move to detector section
*/
         DetectorDef = read_imaging_detector_table(inFile, &ierr);
         if (ierr !=0) {
            printf("error %i opening detector table in %s\n", 
               ierr, argv[1]);
            return 1;
         }
         nRegion = DetectorDef->nregion;
         for (i=0;i<4;i++) DetectorDef->naxis[i][1] = 1; 
         odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
     if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
   }
   outData = make_imaging_fdata_row(odata);
      }  /* iShutter = 1 */
/*
     prepare output data structures
*/

      i = fits_close_file(inFile, &ierr);
/*
          now process the real data
*/
      if (nShutter == 1) {
         printf("   Extracting data from A-chop photometry File %s\n", argv[1]);
         ATarget = minrtsFrameDataReadFitsRow(argv[1], inRow, &ierr);
         if (ierr != 0) {
            printf("Failed to process A-Data: error: %i\n", ierr);
            return 1;
         }
      }
      if (nShutter == 2) {
         printf("   Extracting data from B-chop photometry File %s\n", argv[2]);
         BTarget = minrtsFrameDataReadFitsRow(argv[2], inRow, &ierr);
         if (ierr != 0) {
            printf("Failed to process B-Data: error: %i\n", ierr);
            return 1;
         }
      }
   }   /* loop over shutters */
   nx = ATarget->nx[0];
   ny = ATarget->ny[0];
   nPix = nx*(long)ny;
/*
      if useSky specified, estimate dark from middle rows of closed apertures
      these are region 1 for A- and region 4 for B-
*/
   if (useSky == 1) {
      printf("   Using SKY flux to estimate cross coupling coefficients\n");
      for (iregion=0;iregion<4;iregion++) for (ix=0;ix<nx;ix++) {
         y = 0.;
         for (iy=10;iy<30;iy++) y = y + ATarget->data[0][ix+nx*iy];
         y = y/20.;
         for (iy=0;iy<ny;iy++) ATarget->data[iregion][ix+nx*iy]=
            ATarget->data[iregion][ix+nx*iy] - y;
         y = 0.;
         for (iy=10;iy<30;iy++) y = y + BTarget->data[3][ix+nx*iy];
         y = y/20.;
         for (iy=0;iy<ny;iy++) BTarget->data[iregion][ix+nx*iy]=
            BTarget->data[iregion][ix+nx*iy] - y;
      }
   } else printf("   Using STAR flux to estimate cross coupling coefficients\n");
/*
      rectify images
      For Afiles window 1 (1-based) is noise;
      put window 4 (A-photometry) interpolated to curvature of window 2 -> window 1
      put window 4 (A-photometry) interpolated to curvature of window 3 -> window 4
      put window 2 (A-photometry) interpolated to curvature of window 2 -> window 2
      put window 3 (A-photometry) interpolated to curvature of window 3 -> window 3
      put window 1 (B-photometry) interpolated to curvature of window 2 -> window 1
      put window 1 (B-photometry) interpolated to curvature of window 3 -> window 4
      put window 2 (B-photometry) interpolated to curvature of window 2 -> window 2
      put window 3 (B-photometry) interpolated to curvature of window 3 -> window 3
*/
   scratch = minrtsFrameDataClone(ATarget);
   midiDiffRectXYWindow(scratch, CoordDef, 2, &ierr);
   if (ierr !=0) {
      printf("midiDiffRectXYWindow returned error of %i\n", ierr);
      return 1;
   }
   pixCpy (ATarget->data[0], scratch->data[3], nPix);
   minrtsFrameDataKill(scratch);

   scratch = minrtsFrameDataClone(ATarget);
   midiDiffRectXYWindow(scratch, CoordDef, 3, &ierr);
   if (ierr !=0) {
      printf("midiDiffRectXYWindow returned error of %i\n", ierr);
      return 1;
   }
   pixCpy (ATarget->data[3], scratch->data[3], nPix);
   minrtsFrameDataKill(scratch);

   scratch = minrtsFrameDataClone(BTarget);
   midiDiffRectXYWindow(scratch, CoordDef, 2, &ierr);
   if (ierr !=0) {
      printf("midiDiffRectXYWindow returned error of %i\n", ierr);
      return 1;
   }
   pixCpy (BTarget->data[0], scratch->data[0], nPix);
   minrtsFrameDataKill(scratch);

   scratch = minrtsFrameDataClone(BTarget);
   midiDiffRectXYWindow(scratch, CoordDef, 3, &ierr);
   if (ierr !=0) {
      printf("midiDiffRectXYWindow returned error of %i\n", ierr);
      return 1;
   }
   pixCpy (BTarget->data[3], scratch->data[0], nPix);
   minrtsFrameDataKill(scratch);

#ifdef DEBUG
   minrtsFrameDataWriteFits(ATarget, "arect.fits", &ierr);
   minrtsFrameDataWriteFits(BTarget, "brect.fits", &ierr);
#endif
/*
      create mask for compression
*/
   mask = minrtsFrameDataClone(ATarget);
   for (iregion=0; iregion< 4; iregion++) {
      for (iy=0; iy<ny;iy++) {
         if ((iy< (ny/2-deltaY)) || (iy > (ny/2+deltaY)) )
            for (ix=0;ix<nx;ix++) mask->data[iregion][ix+nx*iy] = 0.;
            else for (ix=0;ix<nx;ix++) mask->data[iregion][ix+nx*iy] = 1.;
      }
   }
   ACompressed = minrtsFrameDataCompress (ATarget, mask, 1);
   BCompressed = minrtsFrameDataCompress (BTarget, mask, 1);
#ifdef DEBUG
   minrtsFrameDataWriteFits(ACompressed, "acompressed.fits", &ierr);
   minrtsFrameDataWriteFits(BCompressed, "bcompressed.fits", &ierr);
#endif
#ifdef SMOOTHX
/*
      box smooth channels
*/
   smWork = (float *)malloc(nx * sizeof(float));
   if (SMOOTHX != 0) {
      for (iregion = 0; iregion<4; iregion++) {
         minrtsBoxSmooth(nx, ACompressed->data[iregion], smWork, SMOOTHX);
         pixCpy (ACompressed->data[iregion], smWork, nx);
         minrtsBoxSmooth(nx, BCompressed->data[iregion], smWork, SMOOTHX);
         pixCpy (BCompressed->data[iregion], smWork, nx);
      }
   }
#endif
/*
      truncate photometry channels on bottom at 1% of peak
*/
  y = 0.;
  for (ix=0;ix<ACompressed->nx[0];ix++) 
     if (ACompressed->data[0][ix] > y) y=ACompressed->data[0][ix];
  y = 0.01* y;
  if (y > 0) for (ix=0;ix<ACompressed->nx[0];ix++) 
     if (ACompressed->data[0][ix] < y) ACompressed->data[0][ix]=y;

  y = 0.;
  for (ix=0;ix<ACompressed->nx[3];ix++) 
     if (ACompressed->data[3][ix] > y) y=ACompressed->data[3][ix];
  y = 0.01* y;
  if (y > 0) for (ix=0;ix<ACompressed->nx[3];ix++) 
     if (ACompressed->data[3][ix] < y) ACompressed->data[3][ix]=y;

  y = 0.;
  for (ix=0;ix<BCompressed->nx[0];ix++) 
     if (BCompressed->data[0][ix] > y) y=BCompressed->data[0][ix];
  y = 0.01* y;
  if (y > 0) for (ix=0;ix<BCompressed->nx[0];ix++) 
     if (BCompressed->data[0][ix] < y) BCompressed->data[0][ix]=y;

  y = 0.;
  for (ix=0;ix<BCompressed->nx[0];ix++) 
     if (BCompressed->data[3][ix] < y) BCompressed->data[3][ix]=y;
  y = 0.01* y;
  if (y > 0) for (ix=0;ix<BCompressed->nx[3];ix++) 
     if (BCompressed->data[3][ix] < y) BCompressed->data[3][ix]=y;
   for (ix=0;ix<ACompressed->nx[1];ix++) {
      outData->data[0][ix] = ACompressed->data[1][ix]/ACompressed->data[3][ix];
      outData->data[1][ix] = ACompressed->data[2][ix]/ACompressed->data[3][ix];
      outData->data[2][ix] = BCompressed->data[1][ix]/BCompressed->data[0][ix];
      outData->data[3][ix] = BCompressed->data[2][ix]/BCompressed->data[0][ix];
   }
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);
/*
      clean up
*/
/*   i = fits_close_file(outFile, &ierr); */
   i = close_imaging_fdata_table(outFile, odata, &ierr); 
   kill_imaging_fdata_row(outData); 
   kill_imaging_detector_table_def(CoordDef);
   minrtsFrameDataKill (ATarget);
   minrtsFrameDataKill (BTarget);
   minrtsFrameDataKill (ACompressed);
   minrtsFrameDataKill (BCompressed);
   minrtsFrameDataKill (mask);
#ifdef SMOOTHX
   free(smWork);
#endif

   return 0;
}
