;
; Copyright 2005, 2006 University of Leiden.
;
; This file is part of MIA+EWS.
;
; MIA+EWS is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; MIA+EWS is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with MIA; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
PRO calDatabase__DEFINE
a1 = {caldatabase, $
revision:0,  $
origin:'',   $
fileName:'', $
typeList:PTR_NEW(),      $; which types in which tables
multiplicity:PTR_NEW(),  $; how many subtables in each table
tableIndex:PTR_NEW(),    $; index to internal tables
tableObj:PTR_NEW()       $;contained tables
}
RETURN
END

PRO calTable__DEFINE
a1 = {calTable, $
type:'',          $ ; table type
final:0B,         $ ; if true I contain data, rather than an array of other tables
header:OBJ_NEW(), $ ; header from my data file
nRows:0L,         $ number of valid rows in myTable
tables:PTR_NEW() $ if single, these objects contain the real data,else subtables
}
RETURN
END

PRO calTable::cleanup
;
   OBJ_DESTROY,self.header
   if (NOT self.final) then $
      if (PTR_VALID(self.tables)) then $
         for i=0,self.nRows-1 do OBJ_DESTROY,(*self.tables)[i]
   PTR_FREE,self.tables
RETURN
END

FUNCTION calTable::init, inFile, typeList, wType
   nType = N_ELEMENTS(wType)
   final = (nType EQ 1)
   if (final) then begin
      self.final = 1
      self.type  = typeList[wType[0]]
      t          = OBJ_NEW('fitstable', inFile, extNu=wType[0]+1)
      self.header= (t->head())->clone()
      self.tables  = PTR_NEW(t->readrows(),/NO_COPY)
      self.nRows = N_ELEMENTS(*self.tables)
      OBJ_DESTROY,t
   endif else if (nType GT 1) then  begin
      self.final = 0
      self.type  = typeList[wType[0]]
      tableArr   = OBJARR(nType)
      for i=0,nType-1 do $
         tableArr[i] = OBJ_NEW('calTable', inFile, typeList, wType[i])
      self.tables= PTR_NEW(tableArr)
      self.nRows = N_ELEMENTS(tableArr)
   endif
RETURN,1
END

FUNCTION calTable::type  & return,self.type &END
FUNCTION calTable::final & return,self.final & END
FUNCTION calTable::nRows & return,self.nRows & END
FUNCTION calTable::tables& return,self.tables&END

FUNCTION calTable::dataPointers
   if (self.final) then RETURN, self.tables $
   else begin
      outPoint = PTRARR(self.nRows)
      for i=0,self.nRows-1 do outPoint[i] = (*self.tables)[i]->dataPointers()
   endelse
RETURN,outPoint
END

FUNCTION calTable::data 
   if (self.final) then RETURN,*self.tables $
   else begin
      outData = PTRARR(self.nRows)
      for i=0,self.nRows-1 do outData[i] = PTR_NEW((*self.tables)[i]->data())
      RETURN,outData
   endelse
END

FUNCTION calTable::getPar, keyword, matches
   if (self.final) then begin
      values = self.header->getPar(keyword, matches)
      RETURN,values
   endif
   values = (*self.tables)[0]->getPar(keyword,matches)
   if (matches EQ 1) then begin
      values = REPLICATE(values,self.nRows)
      for i=1,self.nRows-1 do values[i] = (*self.tables)[i]->getPar(keyword,matches)
   endif else begin
      print,' keyword '+keyword+' not found in table '+self.type
      RETURN,0
   endelse
RETURN, values
END

PRO calTable::replaceData, newData
   if (NOT self.final) then $
      if (PTR_VALID(self.tables)) then $
         for i=0,self.nRows-1 do OBJ_DESTROY,(*self.tables)[i]
   PTR_FREE,self.tables
   self.tables = PTR_NEW(newData)
   self.nRows = N_ELEMENTS(newData)
   self.header->addPar,'NAXIS2', self.nRows
RETURN
END

PRO calTable::appendToFile, outFileName, outFileObj, first, ierr=ierr
  ;
  if (self.final) then begin
;     create internal fitsTable for data
     intTable = OBJ_NEW('fitsTable',(*self.tables)[0], extna=self.type)
     intTable->replaceHeader, self.header, ierr=ierr
     if (first) then begin
        intTable->newFile, outFileName, iErr=iErr 
        outFileObj = intTable->file()
        first = 0
     endif else intTable->appendToFile, outFileObj   ; establish contact with file
     intTable->writeRows, *self.tables  ; write data
     intTable->detach ;     remove contact from file
     OBJ_DESTROY, intTable
  endif else $
      if (PTR_VALID(self.tables)) then $
         for i=0,self.nRows-1 do (*self.tables)[i]->appendToFile,outFile,$
            outFileObj, first, ierr=ierr
RETURN
END

PRO calDatabase::cleanup
;
   PTR_FREE,self.tableIndex
   PTR_FREE,self.typeList
   PTR_FREE,self.multiplicity
   if (PTR_VALID(self.tableObj)) then begin
      for i=0, N_ELEMENTS(*self.tableObj)-1 do OBJ_DESTROY,(*self.tableObj)[i]
      PTR_FREE,self.tableObj
   endif
RETURN
END

FUNCTION calDatabase::init, input
;
   inFile = OBJ_NEW('fitsfile', input,iErr=iErr)
   if (iErr NE 0) then begin
      midiprinterrmsg
      RETURN,0
   endif
   self.tableIndex = PTR_NEW(inFile->index())
   types = (*self.tableIndex).type
   sTypes= types[SORT(types)]
   uTypes = sTypes[UNIQ(sTypes)]
   uTypes = STRUPCASE(STRTRIM(uTypes,2))
   nTypes   = N_ELEMENTS(uTypes)
   self.typeList = PTR_NEW(uTypes)
   self.tableOBJ = PTR_NEW(OBJARR(nTypes))
   mTables  = INTARR(nTypes)
   for i=0,nTypes-1 do mTables[i] = N_ELEMENTS(WHERE(types EQ uTypes[i]))
   self.multiplicity = PTR_NEW(mTables)
   for i=0,nTypes-1 do print,i,uTypes[i], mTables[i],format='(i5,a25,i5)'
   for i=0,ntypes-1 do begin
      wType = WHERE(types EQ uTypes[i])
      (*self.tableOBJ)[i] = OBJ_NEW('calTable', inFile, types, wType)
   endfor
   OBJ_DESTROY,inFile
RETURN,1
END

FUNCTION calDatabase::findTable,type, tt
   tt = MAX(WHERE(STRUPCASE(STRTRIM(type,2)) EQ (*self.typeList)))
   if (tt LT 0) then RETURN,0 ELSE RETURN,1
END

FUNCTION calDatabase::subTableByKeyword, type, keyword, svalue, ierr=ierr
;       find a subTable by type and string keyword, return pointer 
;       to subtable
   ierr = 0
   if (not self->findTable(type, wType)) then begin
      print,'   table of type '+type+' not present in database '
      RETURN, PTR_NEW()
   endif
   tableObj = (*self.tableOBJ)[wType]
   if (tableObj->final()) then begin
      ierr = 1
      print,'   table of type '+type+' has no subtables '
      RETURN, PTR_NEW()
   endif
   nTables = tableObj->nRows()
   sv        = STRTRIM(STRUPCASE(svalue),2)
   tableKeys = tableObj->getPar(keyword)
   tableKeys = STRTRIM(STRUPCASE(tableKeys),2)
   wKey      = MIN(WHERE(sv EQ tableKeys))
   if (wKey LT 0) then begin
      ierr = 1
      print, '   no match for keyword '+keyword+'= '+svalue+' in database '
      RETURN, PTR_NEW()
   endif
   tables = tableObj->tables()
RETURN, (*tables)[wKey]
END

FUNCTION calDatabase::data, type, ierr=ierr
;
   ierr = 0
   if (self->findTable(type,wType)) then RETURN, (*self.tableOBJ)[wType]->data()
   print,' requested type '+type+' not present in database'
   ierr = 1
RETURN, 0
END
FUNCTION calDatabase::dataPointers, type, ierr=ierr
;
   ierr = 0
   if (self->findTable(type,wType)) then RETURN, (*self.tableOBJ)[wType]->dataPointers()
   print,' requested type '+type+' not present in database'
   ierr = 1
RETURN, 0
END

FUNCTION calDatabase::getPar, type, keyword,ierr=ierr
   ierr = 0
   if (self->findTable(type,wType)) then $
      RETURN, (*self.tableOBJ)[wType]->getPar(keyword)
   print,' requested type '+type+' not present in database'
   ierr = 1
RETURN, ''
END

PRO calDatabase::replaceData, type, newData, ierr=ierr
   ierr = 0
   if (self->findTable(type,wType)) then begin
      (*self.tableOBJ)[wType]->replaceData, newData
      RETURN
   endif
   print,' requested type '+type+' not present in database'
   ierr = 1
RETURN
END

PRO calDatabase::addData, type, newData
   ierr = 0
   if (self->findTable(type,wType)) then begin
      tableObj = (*self.tableOBJ)[wType]
      data = [tableObj->data(), newData]
      (*self.tableOBJ)[wType]->replaceData, data
      RETURN
   endif
   print,' requested type '+type+' not present in database'
   ierr = 1

RETURN
END

PRO calDatabase::newFile, outFileName, ierr=ierr
   first = 1
   outFileObj = OBJ_NEW()
   nTypes = N_ELEMENTS(*self.typeList)
   for i=0, nTypes-1 do (*self.tableOBJ)[i]->appendToFile, $
      outFileName, outFileObj, first, ierr=ierr 
   outFileObj->close
   OBJ_DESTROY,outFileObj
RETURN
END


FUNCTION calDatabase::sourceByName, name, sourceData, diamData, photData, specData, ierr=ierr
;
   ierr      = 0
   nFound    = INTARR(4)
   nFound[*] = 0
   if (self->findTable('SOURCES',ii)) then sourceTable = (*self.tableObj)[ii]
   if (self->findTable('DIAMETERS',ii)) then diamTable = (*self.tableObj)[ii]
   if (self->findTable('SPECTROPHOTOMETRY',ii)) then  specTable = (*self.tableObj)[ii]
   if (self->findTable('PHOTOMETRY',ii)) then  photTable = (*self.tableObj)[ii]
;
   sdp = sourceTable->dataPointers()
   sName  = STRTRIM(name[0],2)
   dNames = STRTRIM(REFORM((*sdp).name),2)
   whereSource = MIN(WHERE(sName EQ dNames))
   if (whereSource lt 0) then begin
      print,name,' NOT FOUND IN SOURCE Table '
      ierr = 1
      RETURN,nFound
   endif
   sourceData = (*sdp)[whereSource]
   nFound[0] = 1
;    diameter data
   ddp = diamTable->dataPointers()
   dNames = STRTRIM(REFORM((*ddp).name),2)
   whereSource = WHERE(sName EQ dNames)
   if (MAX(whereSource) GE 0) then  begin
      diamData = (*ddp)[whereSource]
      nFound[1] = N_ELEMENTS(whereSource)
   endif else diamData = 0
;    photometry data
   pdp = photTable->dataPointers()
   dNames = STRTRIM(REFORM((*pdp).name),2)
   whereSource = WHERE(sName EQ dNames)
   if (MAX(whereSource) GE 0) then  begin
      photData = (*pdp)[whereSource]
      nFound[2] = N_ELEMENTS(whereSource)
   endif else photData = 0
;      spectra
   spdp = specTable->dataPointers()
   dNames = STRTRIM(specTable->getpar('name'),2)
   whereSource = WHERE(sName EQ dNames)
   if (MAX(whereSource) GE 0) then  begin
      specData = spdp[whereSource]
      nFound[3] = N_ELEMENTS(whereSource)
   endif else specData = 0
;
RETURN, nFound
END

FUNCTION calDatabase::formatAlias, alias
;      reform alias to "standard format"
;      for now convert to upper case and remove blanks
   fAlias = STRUPCASE(alias)
   salias = STRSPLIT(fAlias,' ',/ex)
   n = N_ELEMENTS(salias)
   fAlias = salias[0]
   for i=1,n-1 do fAlias = fAlias+salias[i]
   b32 = STRING(REPLICATE(32B,32))
   STRPUT,b32,fAlias,0 ; blank fill
RETURN,b32
END

FUNCTION calDatabase::sourceByAlias, alias, sourceData, diamData, photData, specData, ierr=ierr
   ierr      = 0
   nFound    = INTARR(4)
   nFound[*] = 0
   if (self->findTable('ALIASES',ii)) then aliasTable = (*self.tableObj)[ii]
   ap = aliasTable->dataPointers()
   sAlias = STRTRIM(self->formatAlias(alias),2)
   aNames = STRTRIM(REFORM((*ap).alias),2)
   whereAlias = MAX(WHERE(sAlias EQ aNames))
   if (whereAlias LT 0) then begin
      print,' alias '+alias+' not found in alias table'
      ierr = 1
      RETURN, nFound
   endif
   sName = (*ap)[whereAlias].name
   sName = sName[0]
   RETURN,self->sourceByName( sName, sourceData, diamData, photData, specData, ierr=ierr)
END

FUNCTION calDatabase::formatPos, input,ra=ra,dec=dec
;
   sType = SIZE(input)
   sType = sType[1+sType[0]]
   myInp = input
   if (sType EQ 7) then $; string
      myInp = [double(STRSPLIT(myInp,' ',/ext)), 0] ; string input
   nInp  = N_ELEMENTS(myInp)
   if (nInp EQ 1) then posDegree=double(input) else begin  ;scalar=degree
         if(myInp[0] LT 0.) then myInp[1:*] = - myInp[1:*]
            posDegree = double(myInp[0])+ double(myInp[1])/60
            if (nInp GT 2) then posDegree = posDegree+double(myInp[2])/3600 
   endelse
   if (KEYWORD_SET(ra) AND (nInp GT 1)) then posDegree = 15 * posDegree
RETURN, posDegree
END

FUNCTION calDatabase::sourcesNearPosition, ra, dec, sources, $ 
   nmax=nmax,rmax=rmax,ierr=ierr
;
;          positions can be 
;      (1)   scalar = positions in degrees
;      (2)   array of 3 number = hms or dms
;      (3)   string = hms /dms as string (blank separated)
;       nmax is maximum number of sources to be return (default 30)
;       rmax is maximum allowed distance to specified position (degrees)
;       Returns distances in degrees
;       sources contains source records for those found
;
   ierr = 0
   if (NOT KEYWORD_SET(nmax)) then nmax=30 ; max 30 sources
   if (NOT KEYWORD_SET(rmax)) then rmax=1. ; 1 degree
   rmaxrad = rmax/!radeg ; radians
   rm2     = rmaxrad^2
   raRad = self->formatPos(ra,/ra)/!radeg
   decRad= self->formatPos(dec,/dec)/!radeg
;

   if (self->findTable('SOURCES',ii)) then sourceTable = (*self.tableObj)[ii]
   sp   = sourceTable->dataPointers()
   dRA  = REFORM((*sp).raepp)/!radeg - raRad
   sDec = REFORM((*sp).decepp)/!radeg
;  cosR = SIN(sDec)*SIN(decRad) + COS(sDec)*COS(sDec)*COS(dRa) ; exact
   dDec = (sDec - decRad)*cos(decRad)

   rr2 = dRA^2 + dDec^2
   wRad = WHERE(rr2 LE rm2)

   if (MAX(wRad) LT 0) then begin
      ierr    = 1
      sources = 0
      RETURN,-1
   endif

   rr = SQRT(rr2[wRad])
   rSrt = SORT(rr)
   rr = rr[rSrt]*!radeg ; degrees
   wRad = wRad[rSrt]

   nRad = N_ELEMENTS(wRad)
   if (nRad GT nmax) then begin
      rr  =rr[0:nmax-1]
      wRad=wRad[0:nmax-1]
   endif
   sources = (*sp)[wRad]
RETURN, rr
END

FUNCTION calDatabase::sourceFromFile, fitsFile, diamData=diamdata, $
   photData=photData, specData=specData, rad=rad, ierr=ierr
;      see if you can find a calibrator with rad (degrees) of
;      position contained in a fits file
;      if rad is not specified, default is 1 arc minute
;      RETURNS directly the sourceData structure from the nearest source
;      in the call line it also returns the diameter, photometry and spectrophotometry, if found
;      ierr = 0 -> OK; ierr = 1 ; no sources found
   if (NOT KEYWORD_SET(rad)) then myrad = 1./60. $ ; 1 arc minute
      else myrad = rad ;
   ierr = 0
   myFits = STRSPLIT(fitsfile,' ',/ex)
   myFits = myFits[0]
   fits = OBJ_NEW('fitsfile', myFits,ierr=ierr)
   if (ierr NE 0) then begin
      midiprinterrmsg
      RETURN,0
   endif
   head = fits->head()
   fra = head->getPar('RA') ; should be degrees
   fdec= head->getPar('DEC')
   fname = head->getPar('TARG NAME')
   OBJ_DESTROY,head
   OBJ_DESTROY,fits
   rr=self->sourcesNearPosition( fra, fdec, sources, $ 
      nmax=5,rmax=rad,ierr=ierr)
   if (ierr EQ 1 ) then begin
      print, '   NO sources found within ',60.*myrad,' arc min '
      print, '   of target '+fname+' at ra=',fra,' dec=', fdec
      RETURN,0
   endif
   nSources = N_ELEMENTS(rr)
   print,'   FOUND '+ sources.name+' at distance ',60.*rr,' arc min from '+$
      fname,format='(a, f8.2,a)'
   print,'      RETURNING data from source ',sources[0].name
   nData = self->sourceByName(sources[0].name, sourceData, diamData, $
      photData, specData, ierr=ierr)
RETURN, sourceData
END

FUNCTION calDatabase::spectrumFromFile, fitsFile, diamdata, outwave, ierr=ierr
; 
   sourceData=self->sourceFromFile(fitsFile, diamData=diamdata, $
      specData=specData, rad=rad, ierr=ierr)
   if (ierr NE 0) then begin
      print,'    spectrumFromFile source not found in file '+fitsFile
      RETURN,0
   endif
   specData = *(specData[0])
   inwave = 1.e6*specData.wavelength
   influx   = specData.flux
   outwave = oirgetwavelength(fitsFile)
   outFlux = INTERPOL(influx, inwave, outwave,/spline)

RETURN, outFlux
END
FUNCTION calDatabase::bandpassData, bpName, ierr=ierr
    ; data structure for the given bandpass
    ierr = 0
    bp = self->subTableByKeyword('BANDPASS','BAND',bpName,ierr=ierr)
    if (ierr NE 0) then return,0
RETURN, bp->data()
END
