;
; Copyright 2005, 2006 University of Leiden.
;
; This file is part of MIA+EWS.
;
; MIA+EWS is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; MIA+EWS is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with MIA; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
FUNCTION xmlParseFmt, fmt
;        parse the 'format' section of an xml COLUMN code for CREATE_STRUCT
   fLen = STRLEN(fmt)
   type = STRMID(fmt,fLen-1,1)
   if (fLen EQ 1) then begin
      CASE type OF
         'B': outfmt = '0B'
         'L': outfmt = '0B'
         'I': outfmt = '0'
         'E': outfmt = '0.0'
         'D': outfmt = '0.D0'
         'A': outfmt = "''"
      ENDCASE
   endif else begin
      len  = STRMID(fmt,0,fLen-1)
      CASE type OF
         'B': outfmt = 'BYTARR('+len+')'
         'L': outfmt = 'BYTARR('+len+')'
         'I': outfmt = 'INTARR('+len+')'
         'E': outfmt = 'FLTARR('+len+')'
         'D': outfmt = 'DBLARR('+len+')'
         'A': outfmt = "STRING(REPLICATE(32B,"+len+"))"
      ENDCASE
   endelse
   outfmt = outfmt
RETURN, outfmt
END
PRO xmlFormatKeyword, node, unit, last=last
;       format a keyword entry in a __DEFINE procedure
    name  = node->getParm('name')
    deUnderScore,name
    type = STRUPCASE(node->getParm('type'))
    comment = node->getParm('comment')
    deUnderScore,comment
    CASE type OF
       'BOOL':    outfmt = '0B'
       'BYTE':    outfmt = '0B'
       'INTEGER': outfmt = '0'
       'FLOAT':   outfmt = '0.0'
       'DOUBLE':  outfmt = '0.D0'
       'STRING':  outfmt = "''"
    ENDCASE
    if (KEYWORD_SET(last)) then line = '   '+name+':'+outfmt $
       else line = name+':'+outfmt+',' 
    line = line+' $ ; '+comment
    PRINTF,unit,line
RETURN
END

PRO xmlNewKeyword, node, unit
   keyName  = node->getParm('name')
   sDefault = node->getParm('default')
;            assign default value to class variable
   if (sDefault NE '') then printf,unit,"      self."+keyName+"= "+sDefault
;            update class variable if specified in input
   printf,unit,"      if (KEYWORD_SET("+keyName+")) then self."+keyName+"="+keyName
;            copy class variable into FITS header
   printf,unit,"      (self.head)->addPar,'"+keyName+"', self."+keyName
RETURN
END

PRO xmlFormatColumn, node, unit, last=last
;      create CREATE_STRUCT line for this column
    name = node->getParm('name')
    fmt  = STRTRIM(node->getParm('format'),2)
    outFmt = xmlParseFmt(fmt)
    line = "         '"+name+"', "+outFmt
    if (NOT KEYWORD_SET(last)) then printf,unit,line+',$' else $
       printf,unit,line+' )'
RETURN
END

FUNCTION xmlGetNodesByType, xmlSource, type
   subNodes  = xmlSource->getNodes()
   nSubNodes = N_ELEMENTS(subNodes)
   subTypes  = STRARR(nSubNodes)
   for i=0,nSubNodes-1 do subTypes[i] = subNodes[i]->type()
   wType  = WHERE(subTypes EQ STRUPCASE(type))
   if (MAX(wType) LT 0) then RETURN,OBJ_NEW()
RETURN,subNodes[wType]
END

PRO xmlCode__Define, xmlSource, prefix
;       
;      construct and output code for the __DEFINE procedure
   type = xmlSource->type()
   if (type NE 'BINTABLE') then begin
      print,' xmlCodeDefine:: I can only define fits BINTABLES'
      print,'   not '+type
   endif 
   tableName = xmlSource->getParm('name')
   print,'   Constructing IDL BINTABLE __DEFINE for table '+tableName
;
   openw,unit,prefix+tableName+'.pro',/GET_LUN
   comment = xmlSource->getParm('comment')
   deUnderScore,comment
;             get keyword nodes
   keywordNodes = xmlGetNodesByType(xmlSource,'KEYWORD')
   if(OBJ_VALID(keywordNodes[0])) then nKeys = N_ELEMENTS(keywordNodes) $
      else nKeys = 0
   printf,unit,'; Code to access FITS '+tableName+' tables from IDL'
   printf,unit,'; '+comment
   printf,unit,'PRO '+tableName+'__DEFINE'
   printf,unit,'   a1 = {'+tableName+', $ ; tableName '
   printf,unit,'   INHERITS FITSTABLE,  $'
;
   for i=0,nKeys-2 do xmlFormatKeyword,keywordNodes[i],unit
      xmlFormatKeyword,keywordNodes[nKeys-1],unit,/last
   printf,unit,'}'
   printf,unit,'RETURN'
   printf,unit,'END'
   FREE_LUN,unit
RETURN
END

PRO xmlCodeInit, xmlSource, prefix
;          construct the ::init routine for this table
   tableName = xmlSource->getParm('name')
;          get defined keywords from input source
   keywordNodes = xmlGetNodesByType(xmlSource,'KEYWORD')
   if (OBJ_VALID(keywordNodes[0])) then nKeys = N_ELEMENTS(keywordNodes) $
      else nKeys = 0
   keywords = STRARR(nKeys)
   for i=0,nKeys-1 do keywords[i] = STRTRIM(keywordNodes[i]->getParm('name'),2)
;
   print,'   Constructing IDL ::INIT for table '+tableName
;
   openw,unit,prefix+tableName+'.pro',/GET_LUN,/append
;
   printf,unit,' '
;          print FUNCTION definition line and allow user
;          to specify value of keywords at creation
   initLine = "FUNCTION "+tableName+"::init,input, table=table,extNumber=extNumber"
   if (nKeys GT 0) then initLine = initLine + ", $"
   printf,unit,initLine
   for i=0,nKeys-1 do begin
      if (i LT nKeys-1) then printf,unit,"   "+keywords[i]+"="+keywords[i]+",$" $
      else printf,unit,"   "+keywords[i]+"="+keywords[i]
   endfor
   ;determine input type
   printf,unit,'   sInput = size(input)'
   printf,unit,'   sInput = sInput[sInput[0]+1]'
   printf,unit,'   if ((sInput EQ 7) OR (sInput EQ 11)) then begin '
   printf,unit,' ; string input = filename or (FITSFILE) object '
   printf,unit,"      if (0 EQ self->FITSTABLE::init(input[0], "+ $
      "extName='"+STRUPCASE(tableName)+"',$"
   printf,unit,"         extNumber=extNumber,table=table, iErr=iErr)) then begin "
   printf,unit,"         print,'"+tableName+" Creation from Disk Failed '"
   printf,unit,"         RETURN,0"
   printf,unit,'      endif'
;                 initialization from disk succeeded !
;                 now set internal keywords
   for i=0,nKeys-1 do begin
      printf,unit,"      self."+keywords[i]+" = (self.head)->getPar('" +$
         keywords[i]+"', matches)"
   endfor
   printf,unit,'   endif  $; file input '
;                it wasn't file input, create information from scratch
   printf,unit,'   else if (sInput EQ 0) then begin ; no input '
;         format data template
   columnNodes = xmlGetNodesByType(xmlSource,'COLUMN')
   nCols = N_ELEMENTS(columnNodes)
   printf,unit,"      ; format data template "
   printf,unit,"      template=CREATE_STRUCT( $"
   for i=0,nCols-2 do xmlFormatColumn,columnNodes[i], unit
      xmlFormatColumn,columnNodes[nCols-1], unit, /last
;           look for boolean columns
   colType = STRARR(nCols)
   for i=0,nCols-1 do colType(i) = columnNodes[i]->getParm('format')
   logCol  = WHERE(STRTRIM(colType,2) EQ 'L')
   nLogCol = N_ELEMENTS(logCol)
   if (MAX(logCol) GE 0) then begin
      logColS = '         logCol=['
      for i=0,nLogCol-2 do logColS=logColS+STRTRIM(logCol[i],2)+','
      logColS=logColS+STRTRIM(logCol[nLogCol-1],2)+'],$'
   endif else logColS = ''
   printf,unit,";     set me up as a fitsTable"
   printf,unit,"      if (0 EQ self->FITSTABLE::init(template, $"
   if (logColS NE '') then printf,unit,logColS
   printf,unit,"         extname='"+STRUPCASE(tableName)+"', ierr=ierr)) then begin "
   printf,unit,"         print,'"+tableName+" Creation Failed '"
   printf,unit,"         RETURN,0"
   printf,unit,"      endif"
;           put the values of internal variables into FITS header; deal with defaults
   for i=0,nKeys-1 do xmlNewKeyword, keywordNodes[i], unit
   printf,unit,'   endif  else begin $; not recognized input '
   printf,unit,"      print,'Input type not recognized '"
   printf,unit,'      RETURN, 0'
   printf,unit,'   endelse '
   printf,unit,'RETURN, 1'
   printf,unit,'END '
   FREE_LUN,unit
RETURN
END
PRO xmlCodeCleanup, xmlSource, prefix
;          construct the ::cleanup routine for this table
   tableName = xmlSource->getParm('name')
   print,'   Constructing IDL ::CLEANUP for table '+tableName
;
   openw,unit,prefix+tableName+'.pro',/GET_LUN,/append
;
   printf,unit,';'
   printf,unit, "PRO "+tableName+"::cleanup"
   printf,unit, "   self->fitsTable::cleanup"
   printf,unit,'RETURN'
   printf,unit,'END '
   FREE_LUN,unit
RETURN
END
